import numpy as np
import ifigure.utils.geom
from ifigure.interactive import figure, solid
import matplotlib.cm as cm

from .mesh_viewer import open_meshviewer
from petram.helper.geom import connect_pairs

def dim2name_bdry(dim):
    if dim == 3:
        return  'face'
    elif dim == 2:
        return 'edge'
    elif dim == 1:
        return 'point'
    
def dim2name_domain(dim):
    if dim == 3:
        return  'volume'
    elif dim == 2:
        return 'face'
    elif dim == 1:
        return 'edge'


cmap_name = 'Paired'

def get_cmap(name):
    '''
    adjust cmap behavior difference between 1.5 and 2.0
    '''
    m = cm.get_cmap(name)
    data = cm.datad[name]
    m.N = len(data[data.keys()[0]])
    try:
         m(3)
    except IndexError:
         pass
    return m
    
def plot_bdrymesh(meshname = '', mesh = None, idx = 'all',
                  viewer = None, cmap=None, name = '',
                  return_data = False, linewidths = 1.0):
    if mesh is None:  mesh = globals()[meshname]
    if viewer is None:  viewer = open_meshviewer()
    if cmap is None: cmap = cmap_name

    key = dim2name_bdry(mesh.Dimension())
    nbdr = mesh.bdr_attributes.Size()
    kbdr = mesh.bdr_attributes.ToList()
    nbe = mesh.GetNBE()

    
    idxarr = kbdr if str(idx) == 'all' else idx    
    attr = mesh.GetBdrAttributeArray()
    ivert = np.vstack([mesh.GetBdrElement(i).GetVerticesArray() for i in range(nbe)])

    cmap = get_cmap(cmap)
    dom_check_vert = [None]*mesh.bdr_attributes.Size()
    dom_check_idx = [None]*mesh.bdr_attributes.Size()    
    viewer.update(False)

    for kk, ii in enumerate(idxarr):
#        idx = np.where(attr == ii)[0]
#        data = np.dstack([np.vstack([mesh.GetVertexArray(int(k)) for k in ivert[x]]) 
#                          for x in idx])
        idx = mesh.GetBdrArray(ii)
        data = np.dstack([np.vstack([mesh.GetVertexArray(k)
                          for k in mesh.GetBdrElement(x).GetVerticesArray()])
                          for x in idx])

        dom_check_vert[kk] = mesh.GetBdrElementVertices(idx[0])
        dom_check_idx[kk] = idx[0]
        #ivert[attr == ii, :]])
        data = np.rollaxis(data, 2, 0)  ### [i_triangle, i_vert, xyz]
        #data[:,:,2] =         data[:,:,2]*100

        obj= viewer.solid(data, facecolor = cmap(kk % cmap.N),#loat(ii)/nbdr),
#                          edgecolor=(0,0,0,1), alpha = 1.0)
                          alpha = 1.0, linewidths=linewidths)
#                          view_offset = (0.0, 0.0, -0.01, 0.0))
        if obj is not None: obj.rename(name + key + '_'+str(ii))

    viewer.update(True)        
    if return_data: return data
#    return dom_check_vert
    return dom_check_idx


def plot_domainmesh(meshname = '', mesh = None, idx = 'all',
                  viewer = None, cmap=None, name = '',
                    return_data = False, linewidths = 1.0):
    if mesh is None:  mesh = globals()[meshname]
    if viewer is None:  viewer = open_meshviewer()
    if cmap is None: cmap = cmap_name
    
    key = dim2name_domain(mesh.Dimension())    
    ndom = mesh.attributes.Size()
    kdom = mesh.attributes.ToList()
    nele = mesh.GetNE()
    
    idxarr = kdom if str(idx) == 'all' else idx
    
    attr = mesh.GetAttributeArray()
    ivert = np.vstack([mesh.GetElement(i).GetVerticesArray() for i in range(nele)])

    cmap = get_cmap(cmap)
    dom_check_vert = [None]*mesh.attributes.Size()
    dom_check_idx = [None]*mesh.attributes.Size()    
    viewer.update(False)

    for kk, ii in enumerate(idxarr):
        idx = np.where(attr == ii)[0]
        data = np.dstack([np.vstack([mesh.GetVertexArray(k)
                          for k in mesh.GetElement(x).GetVerticesArray()])
                          for x in idx])

        data = np.rollaxis(data, 2, 0)  ### [i_triangle, i_vert, xyz]
        obj= viewer.solid(data, facecolor = cmap(kk % cmap.N), #float(ii)/nele),
#                          edgecolor=(0,0,0,1), alpha = 1.0)
                          alpha = 1.0, linewidths=linewidths)
        if obj is not None: obj.rename(name + key + '_'+str(ii))

    viewer.update(True)        
    if return_data: return data

def find_domain_bdr(mesh, dom_check_list):
    ndom = mesh.attributes.Max()
    ne   = mesh.GetNE()
    kbdr = mesh.bdr_attributes.ToList()

    domainbdr = []
    for i in range(ndom):
        domainbdr.append([])

    '''
    This old method scan all elements...
    for i in range(ndom): domainbdr[i] = list()
    for i in range(ne):
        iattr =  mesh.GetElement(i).GetAttribute()
        x = mesh.GetElementVertices(i)
        idx = np.where([all(y in x for y in t) for t in dom_check_list])[0]
        domainbdr[iattr-1].extend([kbdr[j] for j in idx])
    '''
    for ibdr in dom_check_list:
        idx = mesh.GetBdrElementEdgeIndex(ibdr)
        elem1, elem2 = mesh.GetFaceElements(idx)
        if elem1 != -1:
            i = mesh.GetAttribute(elem1)-1
            domainbdr[i].append(mesh.GetBdrAttribute(ibdr))
        if elem2 != -1:
            i = mesh.GetAttribute(elem2)-1
            domainbdr[i].append(mesh.GetBdrAttribute(ibdr))
            
    return domainbdr
            
def plot_bdr(meshname = '', mesh = None, idx = 'all', viewer = None,
             name = '', return_data = False,  **kwargs):
    if mesh is None:  mesh = globals()[meshname]
    if viewer is None:  viewer = open_meshviewer()

    nbdr = mesh.bdr_attributes.Size()
    kbdr = mesh.bdr_attributes.ToList()
    nbe = mesh.GetNBE()

    idxarr = kbdr if str(idx) == 'all' else idx
    attr = np.array([mesh.GetBdrElement(i).GetAttribute() for i in range(nbe)])
#    ivert = np.vstack([mesh.GetBdrElement(i).GetVerticesArray() for i in range(nbe)])
#    print [x for x in ivert[attr == 1, :]]


    for ii in idxarr:
        idx = np.where(attr == ii)[0]
        edges = np.array([mesh.GetBdrElementEdges(i)[0] for i in idx]).flatten()
        dom_check_vert[ii] = mesh.GetBdrElementVertices(idx[0])
        #ange(nbe)
#                          if mesh.GetBdrElement(i).GetAttribute()==ii]).flatten()
        d = {}
        for x in edges:
            d[x] = x in d
        edges = [x for x in d.keys() if not d[x]]
        ivert = [mesh.GetEdgeVertices(x) for x in edges]
        ivert = connect_pairs(ivert)
        vv = np.vstack([mesh.GetVertexArray(i) for i in ivert])
        obj = viewer.plot(vv[:,0], vv[:,1], vv[:,2], **kwargs)
        if obj is not None: obj.rename(name + 'bdry_'+str(ii))        

    if return_data: return vv
    return dom_check_vert


def plot_domain(meshname = '', mesh = None, idx = 'all',
                viewer = None, cmap=None, return_data = False):
    if mesh is None:  mesh = globals()[meshname]
    if viewer is None:  viewer = open_meshviewer()
    if cmap is None: cmap = cmap_name

    cmap = get_cmap(cmap)    

    nbdr = mesh.bdr_attributes.Size()
    ndom = mesh.attributes.Size()    
    kdom = mesh.attributes.ToList()
    nbe = mesh.GetNBE()
    nel = mesh.GetNE()    

    idxarr = kdom if str(idx) == 'all' else idx    
    dom_attr = np.array([mesh.GetElement(i).GetAttribute() for i in range(nel)])
    bdr_attr = np.array([mesh.GetBdrElement(i).GetAttribute() for i in range(nbe)])
    dfaces = np.array([mesh.GetElementFaces(i)[0] for i in range(nel)])
    bfaces = [mesh.GetBdrElementFace(i)[0] for i in range(nbe)]
    bfaces0 = np.unique(bfaces)
    viewer.update(False)
    for kk, ii in enumerate(idxarr):
        domain_faces = np.unique(np.array([dfaces[dom_attr==ii]]).flatten())
        faces = np.intersect1d(domain_faces, bfaces0)
        bindex = [bfaces.index(f) for f in faces]
        bindex = np.unique(bdr_attr[bindex])
        
        data = np.dstack([np.vstack([mesh.GetVertexArray(x)
                                     for x in mesh.GetFaceVertices(kk)]) for kk in faces])
        data = np.rollaxis(data, 2, 0)  ### [i_triangle, i_vert, xyz]    
        #data[:,:,2] =         data[:,:,2]*100
        col =cmap(kk % cmap.N)#float(ii)/ndom)
        obj= viewer.solid(data, facecolor = col, 
                          linewidth = 0.0)
        if obj is not None: obj.rename('domain_'+str(ii))                
        plot_bdr(mesh = mesh, idx = bindex, viewer = viewer,  color = [0,0,0,1],
                 name = 'domain_'+str(ii)+'_')
    viewer.update(True)        
    if return_data: return data

    



from petram.mfem_config import use_parallel
import numpy as np
import itertools
from scipy.sparse import lil_matrix
import itertools
from collections import defaultdict, OrderedDict
from mfem.common.mpi_debug import nicePrint
from mfem.common.parcsr_extra import ToScipyCoo

from petram.helper.dof_map import get_empty_map

import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('hcurln')

if use_parallel:
    USE_PARALLEL = True
    import mfem.par as mfem
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    myid = comm.rank
    nprc = comm.size
    from mfem.common.mpi_debug import nicePrint, niceCall
    from petram.helper.mpi_recipes import *
else:
    USE_PARALLEL = False
    import mfem.ser as mfem
    myid = 0
    nprc = 1
    mfem.ParFiniteElementSpace = type(None)
    mfem.ParGridfunction = type(None)
    mfem.ParMesh = type(None)
    mfem.ParMixedBilinearForm = type(None)

    def nicePrint(*x):
        print(*x)

rules = {}


def eval_element_center(fe, trans):
    nodes = fe.GetNodes()
    ptx = [trans.Transform(ip) for ip in nodes]

    return np.mean(np.vstack(ptx), 0)


def get_rule(fe1o, fe2, trans, orderinc=1, verbose=True):
    fe2o = fe2.GetOrder()

    order = fe1o + fe2o + trans.OrderW() + orderinc
    if (fe2.Space() == mfem.FunctionSpace.rQk):
        assert False, "not supported"

    if not (fe2.GetGeomType(), order) in rules:
        ir = mfem.IntRules.Get(fe2.GetGeomType(), order)
        rules[(fe2.GetGeomType(), order)] = ir

    ir = rules[(fe2.GetGeomType(), order)]
    if verbose:
        dprint1("Order/N Points: ", order, "/", ir.GetNPoints())
    return ir


ref_shapes = {}
ref_shapes["tri"] = [np.array(x, dtype=float)
                     for x in [[0, 0], [1, 0], [0, 1]]]
ref_shapes["quad"] = [np.array(x, dtype=float)
                      for x in [[0, 0], [1, 0], [1, 1], [0, 1]]]
ref_shapes["tet"] = [np.array(x, dtype=float) for x in
                     [[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]]]
ref_shapes["hex"] = [np.array(x, dtype=float) for x in
                     [[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1],
                      [1, 1, 0], [1, 0, 1], [0, 1, 1], [1, 1, 1]]]


def eval_coeff(coeff, T, ip, MV):
    if isinstance(coeff[0], mfem.MatrixCoefficient):
        coeff[0].Eval(MV[1], T, ip)
        if coeff[1] is None:
            return MV[1].GetDataArray()
        else:
            d = MV[1].GetDataArray().copy()
            coeff[1].Eval(MV[1], T, ip)
            d = d + 1j*MV[1].GetDataArray().copy()
            return d
    elif isinstance(coeff[0], mfem.VectorCoefficient):
        coeff[0].Eval(MV[0], T, ip)
        if coeff[1] is None:
            return MV[0].GetDataArray()
        else:
            d = MV[0].GetDataArray().copy()
            coeff[1].Eval(MV[0], T, ip)
            d = d + 1j*MV[0].GetDataArray().copy()
            return d
    elif isinstance(coeff[0], mfem.Coefficient):
        value = coeff[0].Eval(T, ip)
        if coeff[1] is None:
            return np.array(value)
        else:
            value = value + 1j*coeff[1].Eval(T, ip)
            return np.array(value)
    else:
        assert False, "unsupported type of coefficient: " + str(type(coeff[0]))


def get_inv_doftrans(doftrans, sign1):
    Msign = np.diag(sign1.flatten())
    if doftrans is not None:
        Mdoftrans = np.zeros((doftrans.Size(), doftrans.Size()))
        vv = mfem.Vector(doftrans.Size())
        for i in range(doftrans.Size()):
            vv.Assign(0)
            vv[i] = 1
            doftrans.InvTransformPrimal(vv)
            Mdoftrans[:, i] = vv.GetDataArray()

        #if myid == 1: print(Mdoftrans.astype(int))
        return Mdoftrans.dot(Msign)
    else:
        return Msign


def check_inv_doftrans(doftrans):
    Mdoftrans = np.zeros((doftrans.Size(), doftrans.Size()))
    vv = mfem.Vector(doftrans.Size())
    for i in range(doftrans.Size()):
        vv.Assign(0)
        vv[i] = 1
        doftrans.InvTransformPrimal(vv)
        Mdoftrans[:, i] = vv.GetDataArray()
    print(Mdoftrans)
    Mdoftrans = np.zeros((doftrans.Size(), doftrans.Size()))
    for i in range(doftrans.Size()):
        vv.Assign(0)
        vv[i] = 1
        doftrans.TransformPrimal(vv)
        Mdoftrans[:, i] = vv.GetDataArray()
    print(Mdoftrans)
    Mdoftrans = np.zeros((doftrans.Size(), doftrans.Size()))
    for i in range(doftrans.Size()):
        vv.Assign(0)
        vv[i] = 1
        doftrans.InvTransformDual(vv)
        Mdoftrans[:, i] = vv.GetDataArray()
    print(Mdoftrans)

    Mdoftrans = np.zeros((doftrans.Size(), doftrans.Size()))
    for i in range(doftrans.Size()):
        vv.Assign(0)
        vv[i] = 1
        doftrans.TransformDual(vv)
        Mdoftrans[:, i] = vv.GetDataArray()
    print(Mdoftrans)


def map_ir(fe1, eltrans, coeff1,
           shape1, dim2, sdim2, locnors2,
           sign1, Mdoftrans, MV, th=1e-7):

    g1 = fe1.GetGeomType()
    ip = mfem.IntegrationPoint()

    res = []

    def get_options(g1, *pp):
        if g1 == 2 or g1 == 3:
            s = ref_shapes['tri'] if g1 == 2 else ref_shapes['quad']
            options = [pp[0]*s[i] + (1-pp[0])*s[j] for i, j
                       in itertools.permutations(range(len(s)), 2)]
        elif g1 == 4 or g1 == 5:
            s = ref_shapes['tet'] if g1 == 4 else ref_shapes['hex']
            options = [pp[0]*s[i] + pp[1]*s[j] + (1-pp[0]-pp[1])*s[k] for i, j, k
                       in itertools.permutations(range(len(s)), 3)]
        else:
            print(pp)
            assert False, "geometry type: " + str(g1) + " is not supported."
        return options

    d_misalginment = []

    for data in locnors2:
        pp = data[:dim2]
        p = data[dim2:dim2 + sdim2]
        nor = data[dim2 + sdim2:]

        options = get_options(g1, *pp)

        dd = []

        pxs = []
        for o in options:
            if g1 == 2 or g1 == 3:
                ip.Set2(*o)
            elif g1 == 4 or g1 == 5:
                ip.Set3(*o)
            else:
                assert False, "geometry type: " + \
                    str(g1) + " is not supported."
            px = eltrans.Transform(ip)
            pxs.append(px)
            dd.append(np.sum((p - px)**2))

        assert np.min(dd) < np.max(dd) * \
            th, "point not found: " + str(np.min(dd))

        kk = np.argmin(dd)
        if g1 == 2 or g1 == 3:
            ip.Set2(*options[kk])
        elif g1 == 4 or g1 == 5:
            ip.Set3(*options[kk])
        else:
            assert False, "geometry type: " + str(g1) + " is not supported."
        #if myid == 1: print("location", pxs[kk])
        eltrans.SetIntPoint(ip)
        fe1.CalcVShape(eltrans, shape1)

        # if doftrans is not None:
        s1 = shape1.GetDataArray().transpose().dot(Mdoftrans)
        # else:
        #    s1 = shape1.GetDataArray().transpose().dot(Msign)

        cc = eval_coeff(coeff1, eltrans, ip, MV)

        val = nor.dot(cc.dot(s1))  # *sign1
        # shape1.GetDataArray().transpose())) * sign1   #/ww

        res.append(val)
        d_misalginment.append(np.min(dd))

    return np.vstack(res), np.array(d_misalginment)


def hcurln(fes1, fes2, coeff,
           is_complex=False, bdr='all', orderinc=1, verbose=False):

    mat, rstart = get_empty_map(fes2, fes1, is_complex=is_complex)
    mat2, rstart = get_empty_map(fes2, fes1, is_complex=is_complex)

    from petram.helper.element_map import map_element

    name_fes1 = fes1.FEColl().Name()[:2]
    name_fes2 = fes2.FEColl().Name()[:2]

    if verbose:
        if myid == 0:
            dprint1("fes", name_fes1, name_fes2)

    mesh1 = fes1.GetMesh()
    mesh2 = fes2.GetMesh()

    mesh2.Print("/home/shiraiwa/part.mesh")

    if verbose:
        if myid == 0:
            dprint1("NE", mesh1.GetNE(), mesh2.GetNE())
    elmap, elmap_r = map_element(mesh1, mesh2, bdr, map_bdr=True)

    sdim1 = mesh1.SpaceDimension()
    sdim2 = mesh1.SpaceDimension()
    dim1 = mesh1.Dimension()
    dim2 = mesh2.Dimension()

    shape1 = mfem.DenseMatrix()
    shape2 = mfem.Vector()
    ip = mfem.IntegrationPoint()
    nor = mfem.Vector(sdim1)

    if USE_PARALLEL:
        # this is global TrueDoF (offset is not subtracted)
        P = fes1.Dof_TrueDof_Matrix()
        P1mat = ToScipyCoo(P).tocsr()
        #VDoFtoGTDoF1 = P.indices
        #P = fes2.Dof_TrueDof_Matrix()
        #P = ToScipyCoo(P).tocsr()
        #VDoFtoGTDoF2 = P.indices
        #P2mat = P

    vdofs1_senddata = []

    shared_data = []

    el2_2_node = {}
    el2_2_el1 = {}

    for d in elmap_r:
        for x in list(elmap_r[d]):
            el2_2_node[x] = d
        for x in list(elmap_r[d]):
            el2_2_el1[x] = elmap_r[d][x]

    # working for fes2
    # find boundary element on mesh1 using mesh2 boundary
    el2_arr = [list() for x in range(nprc)]
    el1_arr = [list() for x in range(nprc)]
    fe2o_arr = [list() for x in range(nprc)]
    for i_el in range(fes2.GetNE()):
        attr = fes2.GetAttribute(i_el)
        if bdr != 'all' and not attr in bdr:
            continue
        el1_arr[el2_2_node[i_el]].append(el2_2_el1[i_el])
        el2_arr[el2_2_node[i_el]].append(i_el)
        fe2 = fes2.GetFE(i_el)
        fe2o_arr[el2_2_node[i_el]].append(fe2.GetOrder())

    if USE_PARALLEL:
        el1_arr = alltoall_vector(el1_arr, int)  # transfer to mesh1 owners

    # working for fes1
    # find elemet order on mesh1
    fe1o_arr = [list() for x in range(nprc)]
    i_fe1_arr = [list() for x in range(nprc)]
    rank = 0
    for rank, i_bdrs in enumerate(el1_arr):
        for i_bdr in i_bdrs:
            iface = mesh1.GetBdrElementEdgeIndex(i_bdr)
            transs = mesh1.GetFaceElementTransformations(iface)
            i_el1 = transs.Elem1No
            assert transs.Elem2No == -1, "boundary must be exterior for this operator"
            fe1 = fes1.GetFE(i_el1)
            fe1o_arr[rank].append(fe1.GetOrder())
            i_fe1_arr[rank].append(i_el1)
        rank = rank + 1

    if USE_PARALLEL:
        fe1o_arr = alltoall_vector(fe1o_arr, int)  # transfer to mesh2

    # working for fes2
    locnor_arr = [list() for x in range(nprc)]
    data2_arr = [list() for x in range(nprc)]
    verbose1 = verbose
    for rank, i_el2s in enumerate(el2_arr):
        for i_el2, fe1o in zip(i_el2s, fe1o_arr[rank]):
            eltrans = fes2.GetElementTransformation(i_el2)
            fe2 = fes2.GetFE(i_el2)
            nd2 = fe2.GetDof()
            ir = get_rule(
                fe1o,
                fe2,
                eltrans,
                orderinc=orderinc,
                verbose=verbose1)
            verbose1 = False
            shape2.SetSize(nd2)
            data2 = []
            locnors = []
            for jj in range(ir.GetNPoints()):
                ip1 = ir.IntPoint(jj)
                eltrans.SetIntPoint(ip1)
                w = eltrans.Weight() * ip1.weight
                mfem.CalcOrtho(eltrans.Jacobian(), nor)
                nor2 = nor.GetDataArray() / np.linalg.norm(nor.GetDataArray())
                fe2.CalcShape(ip1, shape2)

                if dim2 == 1:
                    d = np.array([ip1.x] + list(eltrans.Transform(ip1))
                                         + list(nor2))
                    locnors.append(d)
                elif dim2 == 2:
                    d = np.array([ip1.x, ip1.y] + list(eltrans.Transform(ip1))
                                                + list(nor2))

                    locnors.append(d)
                else:
                    assert False, "boundary mesh must be dim=1 or 2"
                data2.append(w * shape2.GetDataArray().copy())

            #   np.vstack(locnors).shape = (#Npoints, dim2+sdim2*2)
            #   np.vstack(data2).shape = (#Npoints, #NDoF2)
            #print("size here", np.vstack(locnors).shape, np.vstack(data2).shape)

            locnor_arr[rank].append(np.vstack(locnors))
            data2_arr[rank].append(np.vstack(data2))

    if USE_PARALLEL:
        locnor_arr = alltoall_vectorv(locnor_arr, float)  # transfer to mesh1

    ll = dim2 + 2 * sdim2

    vdofs1_arr = [list() for x in range(nprc)]
    data1_arr = [list() for x in range(nprc)]

    # space to compute the coefficient
    MV = [mfem.Vector(sdim1),
          mfem.DenseMatrix(sdim1, sdim1)]

    max_misalignment = -np.inf
    for rank, i_fe1s in enumerate(i_fe1_arr):
        locnorss = locnor_arr[rank]

        sign_dict = {}

        for k, i_fe1 in enumerate(i_fe1s):
            fe1 = fes1.GetFE(i_fe1)
            nd1 = fe1.GetDof()
            eltrans = fes1.GetElementTransformation(i_fe1)
            doftrans = fes1.GetElementDofTransformation(i_fe1)
            #ctr = eval_element_center(fe1, eltrans)

            locnors2 = locnorss[k]
            shape1.SetSize(nd1, sdim1)
            vdofs1 = fes1.GetElementVDofs(i_fe1)

            dof_sign1 = np.array([[1 if vv >= 0 else -1
                                   for vv in vdofs1], ])
            vdofs1 = [-1 - x if x < 0 else x for x in vdofs1]

            mat_doftrans = get_inv_doftrans(doftrans, dof_sign1)

            if USE_PARALLEL:
                #  After DofTransformation is introduced we can not use GetGlobalTDofNumber, because
                #  element local DoF could be linked with two TrueDoFs in neighber processes
                #  We construct submatrix of Prolongation to construct element matrix
                #  in TrueDof space

                vv1 = [P1mat.indices[P1mat.indptr[ii]:P1mat.indptr[ii+1]]
                       for ii in vdofs1]
                vv3 = [P1mat.data[P1mat.indptr[ii]:P1mat.indptr[ii+1]]
                       for ii in vdofs1]
                ngtof = np.sum([len(x) for x in vv3])
                sub_p = np.zeros((nd1, ngtof))
                k1 = 0
                k2 = 0
                for gtofs, weights in zip(vv1, vv3):
                    for g, w in zip(gtofs, weights):
                        sub_p[k1, k2] = w
                        k2 = k2 + 1
                    k1 = k1 + 1

                vdofs1 = np.hstack(vv1).flatten()
                mat_doftrans = mat_doftrans.dot(sub_p)

            res, misalignment = map_ir(fe1, eltrans, coeff,
                                       shape1, dim2, sdim2,
                                       locnors2, dof_sign1,
                                       mat_doftrans, MV)

            vdofs1_arr[rank].append(np.array(vdofs1))
            data1_arr[rank].append(res)

            max_misalignment = np.max([max_misalignment, np.max(misalignment)])
            # res.shape = (#Npoints, #DoF1)

    if USE_PARALLEL:
        vdofs1_arr = alltoall_vectorv(vdofs1_arr, int)  # transfer to mesh2
        if is_complex:
            data1_arr = alltoall_vectorv(
                data1_arr, complex)  # transfer to mesh2
        else:
            data1_arr = alltoall_vectorv(data1_arr, float)  # transfer to mesh2
        max_misalignment = np.max(
            MPI.COMM_WORLD.gather(max_misalignment, root=0))
    dprint1("Max misalignment: ", max_misalignment)

    shared_data = []

    for rank, i_el2s in enumerate(el2_arr):
        for k, i_el2 in enumerate(i_el2s):
            vdofs1 = vdofs1_arr[rank][k]

            fe2 = fes2.GetFE(i_el2)
            eltrans2 = fes2.GetElementTransformation(i_el2)
            vdofs2 = fes2.GetElementVDofs(i_el2)
            vdofs2 = [-1 - x if x < 0 else x for x in vdofs2]

            d1 = data1_arr[rank][k]
            d2 = data2_arr[rank][k]

            mm = d2.transpose().dot(d1)

            if USE_PARALLEL:
                # prepare data for not-owned DoFs, which will be shared later
                vdofs22 = [fes2.GetLocalTDofNumber(ii) for ii in vdofs2]
                vdofs22g = [fes2.GetGlobalTDofNumber(ii) for ii in vdofs2]

                kkk = 0
                for v2, v2g in zip(vdofs22, vdofs22g):
                    if v2 < 0:
                        shared_data.append([v2g, mm[kkk, :], vdofs1])
                    kkk = kkk + 1
            else:
                vdofs22 = vdofs2

            for i, ltdof2 in enumerate(vdofs22):
                if ltdof2 < 0:
                    continue
                for j, gtdof1 in enumerate(vdofs1):
                    mat[ltdof2, gtdof1] = mat[ltdof2, gtdof1] + mm[i, j]

    if USE_PARALLEL:
        #nicePrint("shared data", shared_data)
        for source_id in range(nprc):
            data = comm.bcast(shared_data, root=source_id)
            myoffset = fes2.GetMyTDofOffset()
            for v2g, elmat, vdofs1 in data:
                if v2g >= myoffset and v2g < myoffset + mat.shape[0]:
                    i = v2g - myoffset
                    for j, gtdof1 in enumerate(vdofs1):
                        mat[i, gtdof1] = mat[i, gtdof1] + elmat[j]
                    #mat[i, vdofs1] = mat[i, vdofs1] + elmat

    from scipy.sparse import coo_matrix, csr_matrix

    if USE_PARALLEL:
        if is_complex:
            m1 = csr_matrix(mat.real, dtype=float)
            m2 = csr_matrix(mat.imag, dtype=float)
        else:
            m1 = csr_matrix(mat.real, dtype=float)
            m2 = None
        from mfem.common.chypre import CHypreMat

        start_col = fes1.GetMyTDofOffset()
        end_col = fes1.GetMyTDofOffset() + fes1.GetTrueVSize()
        col_starts = [start_col, end_col, mat.shape[1]]
        M = CHypreMat(m1, m2, col_starts=col_starts)
    else:
        from petram.helper.block_matrix import convert_to_ScipyCoo
        M = convert_to_ScipyCoo(coo_matrix(mat, dtype=mat.dtype))

    return M



'''
   extract_refined_mesh_data3

   extract refined mesh data from 3D (=ndim) mesh
'''
import numpy as np
from collections import defaultdict

from petram.mesh.find_edges import find_edges
from petram.mesh.find_vertex import find_vertex
from mfem.ser import GlobGeometryRefiner as GR

def extract_refined_mesh_data3(mesh, refine = None):
    
    ndim = mesh.Dimension()
    sdim = mesh.SpaceDimension()
    
    ivert0 = [mesh.GetBdrElement(i).GetVerticesArray()
                           for i in range(mesh.GetNBE())]
    attrs = mesh.GetBdrAttributeArray()
    nvert = np.array([len(x) for x in ivert0])
    idx3 = np.where(nvert == 3)[0]
    idx4 = np.where(nvert == 4)[0]
    ivert = []; ivert3 = None; ivert4 = None
    iv3 = []; iv4 = []
    iv3p = np.array([], dtype=int); iv4p = np.array([], dtype=int)
    
    from petram.mesh.refined_mfem_geom import get_geom

    ptx = []; ivx3 = None; ptx3 = []
    if len(idx3) != 0:
        base = mesh.GetBdrElementBaseGeometry(idx3[0])
        gt = mesh.GetBdrElementTransformation        
        attr3, ptx3, ivx3, ivxe3, attrx3 =  get_geom(idx3, 3, base, gt, attrs,
                                                     sdim,refine)
        ptx.append(ptx3)
        npt = len(ptx3)//len(idx3)        
        lref = len(ivx3)//len(idx3)
        iv3  = [ivert0[k] for k in idx3]
        seen = defaultdict(int)
        for iiv in ivx3[:lref].flatten():
            seen[iiv] += 1
        xxx = np.sort([kk for kk in seen if seen[kk]==1])
        iv3p = np.hstack([xxx+j*npt for j, k in enumerate(idx3)])
        ivert3 = np.vstack(iv3)
    if len(idx4) != 0:
        base = mesh.GetBdrElementBaseGeometry(idx4[0])
        gt = mesh.GetBdrElementTransformation                
        attr4, ptx4, ivx4, ivxe4, attrx4 =  get_geom(idx4, 4, base, gt, attrs,
                                                     sdim,refine)
        ptx.append(ptx4)
        npt = len(ptx4)//len(idx4)
        lref = len(ivx4)//len(idx4)
        iv4 = [ivert0[k] for k in idx4]        
        seen = defaultdict(int)
        for iiv in ivx4[:lref].flatten():
            seen[iiv] += 1
        xxx = np.sort([kk for kk in seen if seen[kk]==1])
        iv4p = np.hstack([xxx+j*npt for j, k in enumerate(idx4)])+len(ptx3)        
        ivert4 = np.vstack(iv4)
        if ivx3 is not None:
            ivx4 = ivx4 + len(ptx3)       
            ivxe4 = ivxe4 + len(ptx3)        
    ptx_face = np.vstack(ptx)
    
    # unique edges for node                       
    tmp = np.hstack(iv3 + iv4)
    tmp2 = np.hstack((iv3p, iv4p))    
    u, findex, indices = np.unique(tmp, return_index=True,
                                   return_inverse=True)
    
    ll3 = 3*len(idx3)
    indices3 = indices[:ll3]
    indices4 = indices[ll3:]
                       
    table = np.zeros(np.max(u)+1, dtype=int)-1
    for u0, fff in zip(u, findex):
        table[u0] = tmp2[fff]
    #for k, u0 in enumerate(u):
    #    table[u0] = tmp2[np.where(tmp == u0)[0][0]]

    X = ptx_face
    #X = np.vstack([mesh.GetVertexArray(k) for k in u])    
    if X.shape[1] == 2:
        X = np.hstack((X, np.zeros((X.shape[0],1))))
    elif X.shape[1] == 1:
        X = np.hstack((X, np.zeros((X.shape[0],2))))
    
    kdom = mesh.attributes.ToList()
    
    cells = {}
    cell_data = {}
    cell_data['X_refined_face']=ptx_face
    if ivert3 is not None:
        cells['triangle'] = indices3.reshape(ivert3.shape)
        cells['triangle_x'] = ivx3
        cells['triangle_xe'] = ivxe3                
        cell_data['triangle'] = {}
        cell_data['triangle']['physical'] = attr3
        cell_data['triangle_x'] = {}                         
        cell_data['triangle_x']['physical'] = attrx3
    if ivert4 is not None:
        cells['quad'] = indices4.reshape(ivert4.shape)
        cells['quad_x'] = ivx4
        cells['quad_xe'] = ivxe4                
        cell_data['quad'] = {}
        cell_data['quad']['physical'] = attr4
        cell_data['quad_x'] = {}                         
        cell_data['quad_x']['physical'] = attrx4


    from petram.mesh.mesh_utils import populate_plotdata
    l_s_loop = populate_plotdata(mesh, table, cells, cell_data)

    ec = mesh.extended_connectivity    
    l2e = ec['line2edge']
    
    ### process refined edges
    # all mfem edge index
    if len(l2e.keys()) > 0:
        idx2 = np.hstack([l2e[key] for key in l2e])
        attr22 = cell_data['line']['physical']
        attr2 = {i:attr22[k]  for k, i in enumerate(idx2)}    

        base = 1
        gt = mesh.GetEdgeTransformation
        attr2, ptx2, ivx2, ivxe2, attrx2 = get_geom(idx2, 2, base, gt, attr2,
                                                sdim, refine)

        cells['line_x'] = ivx2
        cell_data['line_x'] = {}                        
        cell_data['line_x']['physical'] = attrx2
        cell_data['X_refined_edge']=ptx2

    iedge2bb = None # is it used?        
    #print "X", X.shape
    #for k in cells['vertex']:print X[k]
    return X, cells, cell_data, l_s_loop, iedge2bb


    '''
    ## fill line/surface loop info
    loop= {}
    for k in kdom:
        loop[k] = []
        
    for ibdr in range(mesh.GetNBE()):
        idx = mesh.GetBdrElementEdgeIndex(ibdr)
        elem1, elem2 = mesh.GetFaceElements(idx)
        if elem1 != -1:
            i = mesh.GetAttribute(elem1)
            loop[i].append(attrs[ibdr])
        if elem2 != -1:
            i = mesh.GetAttribute(elem2)
            loop[i].append(attrs[ibdr])
    for k in kdom:
        loop[k] = np.unique(loop[k])
        
    l_s_loop = [None, loop]
        
    ## fill line

    edges, bb_edges = find_edges(mesh) ## bb_edges : face pair -> mfem edge id
    bb_keys = bb_edges.keys()
    kedge = []
    cell_line = []
    iedge2bb = {} ##iedge2bb  edge id (petra) to the face pair tuple
    ll = defaultdict(list);  ## face id (mfem) -> edge id (petra) (line loop)

    for k, key in enumerate(bb_keys):
        iedge2bb[k+1] = key
        for kk in key:
           ll[kk].append(k+1)

    ll.default_factory = None
    l_s_loop[0] = ll
    
    line2edge = {}
    vert2vert = {}
    if len(bb_keys) == 0:
        mesh.extended_connectivity = {'line2vert':None,
                                      'surf2line':l_s_loop[0],
                                      'vol2surf': l_s_loop[1],
                                      'line2edge':line2edge,
                                      'vert2vert':vert2vert}
        
        ## iedge2bb : mapping from edge_id to boundary numbrer set
        ## X, cells, cell_data : the same data strucutre as pygmsh
        return X, cells, cell_data, l_s_loop, iedge2bb

    # all mfem edge index
    idx2 = [ie for key in bb_keys for ie in bb_edges[key]]
    line2edge = {idx+1: bb_edges[key] for idx, key in enumerate(bb_edges)}
    
    attr22 = np.hstack([[k+1]*len(bb_edges[key]) for k, key in enumerate(bb_keys)])

    attr2 = {i:attr22[k]  for k, i in enumerate(idx2)}
    iv2 = [mesh.GetEdgeVertices(ie) for key in bb_keys for ie in bb_edges[key]]
    
    base = 1
    gt = mesh.GetEdgeTransformation
    attr2, ptx2, ivx2, ivxe2, attrx2 = get_geom(idx2, 2, base, gt, attr2,
                                                sdim, refine)

    cells['line'] = table[np.vstack(iv2)]
    cell_data['line'] = {}                
    cell_data['line']['physical'] = attr2
        
    cells['line_x'] = ivx2
    cell_data['line_x'] = {}                        
    cell_data['line_x']['physical'] = attrx2
    cell_data['X_refined_edge']=ptx2
        
    cell_data['vertex'] = {}    
    corners, iverts = find_vertex(mesh, bb_edges)
    if len(iverts) != 0:        
        cells['vertex'] = table[iverts]
        cell_data['vertex']['physical'] = np.arange(len(iverts))+1
    
    line2vert = {key: np.array(corners[iedge2bb[key]])+1 for key in iedge2bb}
    vert2vert = {i+1: iverts[i] for i in range(len(iverts))}
    
    mesh.extended_connectivity = {'line2vert':line2vert,
                                  'surf2line':l_s_loop[0],
                                  'vol2surf': l_s_loop[1],
                                  'line2edge':line2edge,
                                  'vert2vert':vert2vert}

    ## iedge2bb : mapping from edge_id to boundary numbrer set
    ## X, cells, cell_data : the same data strucutre as pygmsh
    return X, cells, cell_data, l_s_loop, iedge2bb
    '''



from petram.solver.mumps.hypre_to_mumps import PyZMatrix, PyDMatrix
import numpy as np
from mfem.common.chypre import *

import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('CHypre_2_PyMatrix')


'''
  [ A    B ][x]   [b]
  [        ][ ] = [ ]
  [ C    D ][l]   [c]

  B:  Vertical Vector/Matrix
  C:  Horizontal Vector/Matrix
'''

def Vec2MatH(v, complex=True): # horizontal vector
    vv =  CHypreVec2Array(v)
    if vv is None: return # other than root
    
    if complex:
       m1 = PyZMatrix(1, len(vv), len(vv))
       m1.set_rdata(vv.real)
       m1.set_idata(vv.imag)
    else:
       m1 = PyDMatrix(1, len(vv), len(vv))
       m1.set_data(vv)
    return m1

def Vec2MatV(v, complex=True):  # vertical vector
    vv =  CHypreVec2Array(v)
    if vv is None: return # other than root
    
    if complex:
       m1 = PyZMatrix(len(vv), 1, len(vv))
       m1.set_rdata(vv.real)
       m1.set_idata(vv.imag)
    else:
       m1 = PyDMatrix(len(vv), 1, len(vv))
       m1.set_data(vv)
    return m1

'''
following two are convertion to PyMatrix
'''
def Mat2CooH(m, complex = True, elimination = True): # horizontal matrix  (eliminate zero rows)
    from mpi4py import MPI
    comm     = MPI.COMM_WORLD
    myid     = MPI.COMM_WORLD.rank    

    N = m.shape[1]
    coo = CHypreMat2Coo(m)
    csr = coo.tocsr()
    num_nonzeros = np.diff(csr.indptr)
    k_nonzeros = np.where(num_nonzeros != 0)[0]
    
    if elimination : 
        rr = np.array(comm.allgather(len(k_nonzeros)))
    else:
        rr = np.array(comm.allgather(coo.shape[0]))
    rall = int(np.sum(rr))
    rstart = np.hstack((0, np.cumsum(rr)))

    if np.sum(num_nonzeros != 0) != 0:
        if elimination: csr = csr[np.where(num_nonzeros != 0)[0]]
        nnz = int(csr.getnnz())
        coo = csr.tocoo()
        if complex:
            m2 = PyZMatrix(rall, N, nnz, cmajor=2)
        else:
            m2 = PyDMatrix(rall, N, nnz, cmajor=2)            
        m2.set_irn(coo.row + rstart[myid])
        m2.set_jcn(coo.col)
        if complex:        
            m2.set_rdata(coo.data.real)        
            m2.set_idata(coo.data.imag)
        else:
            m2.set_data(coo.data.astype('float'))                
    else:
        if complex:
            m2 = PyZMatrix(rall, N,  0, cmajor=2)
        else:
            m2 = PyDMatrix(rall, N,  0, cmajor=2)
    return m2

def Mat2CooV(m, complex = True, elimination = True): # vertical offset  (eliminate zero cols)
    '''
    In order to eliminate col, it works on tranposed matrix,
    '''
    from mpi4py import MPI
    comm     = MPI.COMM_WORLD
    myid     = MPI.COMM_WORLD.rank
    
    M = m.shape[0]
    coo = CHypreMat2Coo(m.transpose())

    csr = coo.tocsr()
    
    num_nonzeros = np.diff(csr.indptr)
    k_nonzeros = np.where(num_nonzeros != 0)[0]

    if elimination : 
        rr = np.array(comm.allgather(len(k_nonzeros)))
    else:
        rr = np.array(comm.allgather(coo.shape[0]))
    rall = int(np.sum(rr))
    rstart = np.hstack((0, np.cumsum(rr)))

    if np.sum(num_nonzeros != 0) != 0:
        if elimination: csr = csr[np.where(num_nonzeros != 0)[0]]
        nnz = int(csr.getnnz())
        coo = csr.tocoo()
        if complex:
            m2 = PyZMatrix(M, rall,  nnz, cmajor=2)
        else:
            m2 = PyDMatrix(M, rall,  nnz, cmajor=2)            
        m2.set_jcn(coo.row + rstart[myid])
        m2.set_irn(coo.col)
        
        if complex:        
            m2.set_rdata(coo.data.real)        
            m2.set_idata(coo.data.imag)
        else:
            m2.set_data(coo.data.astype('float'))                
    else:
        if complex:
            m2 = PyZMatrix(M, rall,  0, cmajor=2)
        else:
            m2 = PyDMatrix(M, rall, 0, cmajor=2)
    return m2    


def NonZeroCol(m):
    from mpi4py import MPI
    comm     = MPI.COMM_WORLD
    myid     = MPI.COMM_WORLD.rank    

    N = m.shape[1]
    coo = CHypreMat2Coo(m.transpose())
    csr = coo.tocsr()
    num_zeros = np.diff(csr.indptr)
    k_zeros = np.where(num_zeros == 0)[0]
    dprint1("ZeroCol", coo.shape)
    rr = np.array(comm.allgather(len(k_zeros)))
    return rr
    
def allgather_vector(data, mpi_data_type = None, assemble=False):
    from mpi4py import MPI       
    from mfem.common.mpi_dtype import  get_mpi_datatype       
    if mpi_data_type is None:
         mpi_data_type = get_mpi_datatype(data)

    myid     = MPI.COMM_WORLD.rank
    rcounts = data.shape[0]
    rcounts = np.array(MPI.COMM_WORLD.allgather(rcounts))
    for x in data.shape[1:]: rcounts = rcounts * x        
    cm = np.hstack((0, np.cumsum(rcounts)))
    disps = list(cm[:-1])        
    length =  cm[-1]
    recvbuf = np.empty([length], dtype=data.dtype)
    recvdata = [recvbuf, rcounts, disps, mpi_data_type]
    senddata = [data.flatten(), data.flatten().shape[0]]        
    MPI.COMM_WORLD.Allgatherv(senddata, recvdata)
    return recvbuf.reshape(-1, *data.shape[1:])

def ZeroColIndex(m):
    from mpi4py import MPI
    comm     = MPI.COMM_WORLD
    myid     = MPI.COMM_WORLD.rank    

    N = m.shape[1]
    coo = CHypreMat2Coo(m.transpose())        
    rr = np.array(comm.allgather(coo.shape[0]))
    rall = int(np.sum(rr))
    rstart = np.hstack((0, np.cumsum(rr)))

    csr = coo.tocsr()
    num_zeros = np.diff(csr.indptr)
    k_zeros = np.where(num_zeros == 0)[0] + rstart[myid]
    #dprint1("ZeroCol", coo.shape, k_zeros)
    if MPI.INT.size == 4:
        dtype = np.int32
    else:
        dtype = np.int64
    k_zeros_all = allgather_vector(k_zeros.astype(dtype), MPI.INT, False)
    return k_zeros_all




from __future__ import print_function

import os
import six
import numpy as np


class Solfiles(object):
    '''
    solfile list container ( used to refer it weakly)
    '''

    def __init__(self, l, no_scan=False):
        if isinstance(l, Solfiles):
            #print("in this case")
            self.set = l.set
            self.parametric_data = l.parametric_data
        else:
            #print("in that case")
            self.set = l
            if not no_scan:
                self.check_parametric_data()

    def __len__(self):
        return len(self.set)

    def __getitem__(self, idx):
        items = Solfiles(self.set[idx], no_scan=True)
        items.parametric_data = self.parametric_data
        print("parametric data", items.parametric_data)
        return items

    @property
    def path(self):
        return os.path.dirname(self.set[0][0][0])

    def store_timestamps(self):
        self.timestamps = {}
        for meshes, solf, in self.set:
            for x in meshes:
                self.timestamps[x] = os.path.getmtime(x)

            for key in solf:
                fr, fi = solf[key]
                if fr is not None:
                    self.timestamps[fr] = os.path.getmtime(fr)
                if fi is not None:
                    self.timestamps[fi] = os.path.getmtime(fi)

    def is_different_timestamps(self, solfiles):
        for x in self.timestamps:
            if not x in solfiles.timestamps:
                return True
            if solfiles.timestamps[x] != self.timestamps[x]:
                return True
        return False

    @property
    def has_parametic_data(self):
        return len(self.parametric_data) > 0

    def check_parametric_data(self):
        import petram.helper.pickle_wrapper as pickle
        self.parametric_data = []

        bname = os.path.basename(self.path)
        idx = -1
        try:
            if bname.startswith('case_'):
                idx = int(bname.split('_')[-1])
        except BaseException:
            return

        if idx == -1:
            return

        for f in os.listdir(os.path.dirname(self.path)):
            if f.startswith("parametric_data"):
                try:
                    params = pickle.load(open(f, 'rb'))
                    self.parametric_data.append(
                        (params['name'], params['data'][idx]))
                except BaseException:
                    pass
        print("parametric data", self.parametric_data)


class MeshDict(dict):
    pass


class Solsets(object):
    '''
    Solsets: bundle of GridFunctions

      methes: names, meshes, gfr, gfi
    '''

    def __init__(self, solfiles, refine=0):
        def fname2idx(t):
            i = int(os.path.basename(t).split('.')[0].split('_')[-1])
            return i
        solfiles = solfiles.set
        object.__init__(self)
        self.set = []
        import mfem.ser as mfem

        fix_orientation = False  #false
        generate_edge = 1       #1
        refine = 0              #1
        
        for meshes, solf, in solfiles:
            idx = [fname2idx(x) for x in meshes]
            meshes = {
                i: mfem.Mesh(
                    str(x),
                    generate_edge,
                    refine,
                    fix_orientation) for i, x in zip(
                    idx, meshes)}
            meshes = MeshDict(meshes)  # to make dict weakref-able
            # what is this refine = 0 !?
            for i in idx:
                # meshes[i].ReorientTetMesh()
                meshes[i]._emesh_idx = i
            s = {}
            for key in six.iterkeys(solf):
                fr, fi = solf[key]
                i = fname2idx(fr)
                m = meshes[i]

                solr = (mfem.GridFunction(m, str(fr))
                        if fr is not None else None)
                soli = (mfem.GridFunction(m, str(fi))
                        if fi is not None else None)
                if solr is not None:
                    solr._emesh_idx = i
                if soli is not None:
                    soli._emesh_idx = i

                s[key] = (solr, soli)
            self.set.append((meshes, s))

    def __len__(self):
        return len(self.set)

    def __iter__(self):
        return self.set.__iter__()

    @property
    def meshes(self):
        return [x[0] for x in self.set]

    @property
    def names(self):
        ret = []
        for x in self.set:
            ret.extend(x[1].keys())
        return tuple(set(ret))

    def gfr(self, name):
        return [x[1][name][0] for x in self.set]

    def gfi(self, name):
        return [x[1][name][1] for x in self.set]


def find_solfiles(path, idx=None):
    import os

    files = os.listdir(path)
    mfiles = [x for x in files if x.startswith('solmesh')]
    if len(mfiles) == 0:
        return None
    solrfile = [x for x in files if x.startswith('solr')]
    solifile = [x for x in files if x.startswith('soli')]

    if len(mfiles) == 0:
        '''
        mesh file may exist one above...shared among parametric
        scan...
        '''
        files2 = os.listdir(os.path.dirname(path))
        mfiles = [x for x in files2 if x.startswith('solmesh')]
        pathm = os.path.dirname(path)
    else:
        pathm = path

    solfiles = []
    x = mfiles[0]
    suffix_list = list(set(['' if len(x.split('.')) == 1 else '.' + x.split('.')[-1]
                            for x in mfiles]))
    for s in suffix_list:
        meshes = [x for x in mfiles if (len(x.split('.')) == 1 and s == '') or
                  x.endswith(s)]
        meshes = [os.path.join(pathm, x) for x in meshes]
        solrs = [x for x in solrfile if (len(x.split('.')) == 1 and s == '') or
                 x.endswith(s)]
        solis = [x for x in solifile if (len(x.split('.')) == 1 and s == '') or
                 x.endswith(s)]
        names = ['_'.join(x.split('.')[0].split('_')[1:]) for x in solrs]

        sol = {}
        for n in names:
            solr = (os.path.join(path, 'solr_' + n + s)
                    if ('solr_' + n + s) in solrfile else None)
            soli = (os.path.join(path, 'soli_' + n + s)
                    if ('soli_' + n + s) in solifile else None)

            if solr is None:
                continue
            sol[n] = (solr, soli)
        solfiles.append([meshes, sol])

    ret = Solfiles(solfiles)

    ret.store_timestamps()

    return ret


def read_solsets(path, idx=None, refine=0):
    solfiles = find_solfiles(path, idx)
    return Solsets(solfiles, refine=refine)


read_sol = read_solsets



'''

   MPI helper

   NOTE: do NOT Load mpi4py in global space in this module
'''
from mfem.common.mpi_debug import nicePrint, niceCall


def nicePrint1(*s):
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
    nproc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank

    if myid == 0:
        print("[ID:0]" + ' '.join([str(ss) for ss in s]))



'''

   evaluators

    This module provides objects to evaluate solutions
    in various ways. 


      BdrNodalEvaluator : evaulation on boundary surfaces

         eval:
         evalinteg:

      (plan)
      NodalEvaluator : evaulation on volume

         eval:
         evalinteg:


'''

import numpy as np
import six
import weakref
import os
from weakref import WeakKeyDictionary as WKD
from weakref import WeakValueDictionary as WVD


from petram.mfem_config import use_parallel
if use_parallel:
    import mfem.par as mfem
else:
    import mfem.ser as mfem


def evaluator_cls():
    from petram.sol.bdr_nodal_evaluator import BdrNodalEvaluator
    from petram.sol.slice_evaluator import SliceEvaluator
    from petram.sol.edge_nodal_evaluator import EdgeNodalEvaluator
    from petram.sol.ncface_evaluator import NCFaceEvaluator
    from petram.sol.ncedge_evaluator import NCEdgeEvaluator
    from petram.sol.probe_evaluator import ProbeEvaluator
    from petram.sol.pointcloud_evaluator import PointcloudEvaluator
    from petram.sol.integral_evaluator import IntegralEvaluator

    return {'BdrNodal': BdrNodalEvaluator,
            'EdgeNodal': EdgeNodalEvaluator,
            'NCFace':   NCFaceEvaluator,
            'NCEdge':   NCEdgeEvaluator,
            'Slice': SliceEvaluator,
            'Pointcloud': PointcloudEvaluator,
            'Integral': IntegralEvaluator,
            'Probe': ProbeEvaluator, }


class Evaluator(object):
    '''
    Evaluator is a base class. Mostly empty.
    '''

    def __init__(self):
        object.__init__(self)
        self.failed = False

    def make_agents(self, name, params, **kwargs):
        self._agent_params = (name, params)
        self._agent_kwargs = kwargs.copy()

    '''
    check if evauators are build with current need
    '''

    def validate_param(self, params):
        return self._agent_params[1] == params

    def validate_cls(self, name):
        return self._agent_params[0] == name

    def validate_evaluator(self, name, params, *args, **kwargs):
        val = self.validate_param(params) and self.validate_cls(name)
        if not val:
            return False
        # print 'comparing', self._agent_kwargs, kwargs

        for key in six.iterkeys(kwargs):
            if not key in self._agent_kwargs:
                return False

            # print("comparing")
            # print(kwargs[key],self._agent_kwargs[key])
            #print(kwargs[key] != self._agent_kwargs[key])

            chk = (kwargs[key] != self._agent_kwargs[key])
            if isinstance(chk, bool):
                if chk:
                    return False
            else:
                if chk.any():
                    return False
        return True

    def set_model(self, model):
        '''
        single
        MP:    save file in a temprary directoy and send signal
        server: save file, transfer it, send signal
        '''
        raise NotImplementedError('subclass needs to implement this')

    def set_solfiles(self, solfiles):
        '''
        single
        MP:     send signal
        server: send signal
        '''
        raise NotImplementedError('subclass needs to implement this')

    def check_sol_new(self, client_status):
        '''
        return True if it needs to load model file
        on single : use client_status
        on MP     : use client_status
        on server : time stamp of solfile
        '''
        raise NotImplementedError('subclass needs to implement this')

    def load_solfiles(self, mfem_mode=None):
        raise NotImplementedError('subclass needs to implement this')

    def set_phys_path(self, phys_path):
        raise NotImplementedError('subclass needs to implement this')

    def eval(self, expr):
        raise NotImplementedError('subclass needs to implement this')

    def make_probe_agents(self, *args, **kwargs):
        raise NotImplementedError('subclass needs to implement this')


class EvaluatorCommon(Evaluator):
    '''
    EvaluatorCommon provide pertial implementation of methods
    '''

    def set_model(self, model):
        self.mfem_model = weakref.ref(model)

    def load_solfiles(self, mfem_mode=None):
        if self.solfiles is None:
            return

        if self.solfiles in self.solvars:
            return self.solvars[self.solfiles]

        from petram.sol.solsets import Solsets

        solsets = Solsets(self.solfiles)

        print("reading sol variables")
        if self.solfiles.has_parametic_data:
            self.set_parametric_data()

        phys_root = self.mfem_model()["Phys"]
        solvars = phys_root.make_solvars(solsets.set)

        # Setting _emesh_idx from emesh_idx in GridFunction
        for phys in phys_root:
            vnames = phys_root[phys].dep_vars
            gf_var = None
            for name in vnames:
                if name not in solvars[0]:
                    continue
                gf_var = solvars[0][name]
                break

            if gf_var is not None and hasattr(gf_var, "deriv_args"):
                gf_real, gf_imag = gf_var.deriv_args
                eidx = gf_real._emesh_idx if gf_real is not None else gf_imag._emesh_idx
                phys_root[phys]._emesh_idx = eidx
            else:
                # For this variable, correnspoinding GridFunction does not exists.
                # This happens when solution does not exists for a particular variable.
                # We are not plotting this anyway. But we need to set something
                # instead of stopping it here.
                phys_root[phys]._emesh_idx = -1

        pp_root = self.mfem_model()["PostProcess"]
        solvars = pp_root.add_solvars(solsets.set, solvars)
        print("=== solvars= ===", solvars[0])

        self.solvars[self.solfiles] = solvars

        for key in six.iterkeys(self.agents):
            for a, vv in zip(self.agents[key], solsets):
                a.forget_knowns()
        self.solsets = solsets
        return solvars

    def set_parametric_data(self):
        model = self.mfem_model()
        for name, data in self.solfiles.parametric_data:
            parametric = model[name]
            scanner = parametric.get_scanner(nosave=True)
            solvers = parametric.set_scanner_physmodel(scanner)
            scanner.apply_param(data)

    def make_agents(self, name, params, **kwargs):
        if name == 'Probe':
            self.make_probe_agents(name, params, **kwargs)
        elif name == 'Pointcloud':
            self.make_pointcloud_agents(name, params, **kwargs)
        elif name == 'Integral':
            self.make_integral_agents(name, params, **kwargs)
        else:
            self.make_nodalslice_agents(name, params, **kwargs)

    def make_nodalslice_agents(self, name, params, **kwargs):
        print("making new agents", name, params, kwargs)
        super(EvaluatorCommon, self).make_agents(name, params, **kwargs)
        self.agents = {}
        cls = evaluator_cls()[name]
        solsets = self.solsets
        for param in params:
            self.agents[param] = []
            for m, void in solsets:
                a = cls([param], **kwargs)
                a.set_mesh(m)
                self.agents[param].append(a)

    def make_pointcloud_agents(self, name, params, **kwargs):
        print("making new pc agents", name, params, kwargs)
        super(EvaluatorCommon, self).make_agents(name, params, **kwargs)
        self.agents = {}
        cls = evaluator_cls()[name]
        solsets = self.solsets

        params = tuple(params)
        self.agents[params] = []

        for m, void in solsets:
            a = cls(params, **kwargs)
            a.set_mesh(m)
            self.agents[params].append(a)

    def make_integral_agents(self, name, params, **kwargs):
        print("making new integral agents", name, params, kwargs)
        super(EvaluatorCommon, self).make_agents(name, params, **kwargs)
        self.agents = {}
        cls = evaluator_cls()[name]
        solsets = self.solsets

        params = tuple(params)
        self.agents[params] = []

        for m, void in solsets:
            a = cls(params, **kwargs)
            a.set_mesh(m)
            self.agents[params].append(a)

    def make_probe_agents(self, name, params, **kwargs):
        print("making new probe agents", name, params, kwargs)
        super(EvaluatorCommon, self).make_agents(name, params, **kwargs)
        self.agents = {}
        cls = evaluator_cls()[name]
        for param in params:
            self.agents[param] = [cls([param], **kwargs)]

    def eval_probe(self, expr, **kwargs):
        raise NotImplementedError('subclass needs to implement this')
        '''
        phys_path = self.phys_path
        phys = self.mfem_model()[phys_path]
        solvars = self.load_solfiles()
        
        if solvars is None: return None, None
        '''

    def terminate_all(self):
        pass


def_config = {'use_mp': False,
              'use_cs': False,
              'mp_worker': 2,
              'mp_debug': False,
              'cs_worker': 4,
              'cs_server': 'localhost',
              'cs_soldir': '',
              'cs_solsubdir': '',
              'cs_user': ''}


def build_evaluator(params,
                    mfem_model,
                    solfiles,
                    name='BdrNodal',
                    config=def_config,
                    **kwargs):

    from petram.sol.evaluator_mp import EvaluatorMP
    from petram.sol.evaluator_single import EvaluatorSingle
    from petram.sol.evaluator_cs import EvaluatorClient

    from petram.debug import debug_evaluator_mp
    config["mp_debug"] = debug_evaluator_mp

    print("making new evaluator", config, params)

    if not config['use_mp'] and not config['use_cs']:
        evaluator = EvaluatorSingle()
    elif config['use_mp']:
        logfile = 'log' if config['mp_debug'] else False
        evaluator = EvaluatorMP(nproc=config['mp_worker'],
                                logfile=logfile)
    elif config['use_cs']:
        solpath = os.path.join(config['cs_soldir'],
                               config['cs_solsubdir'])
        evaluator = EvaluatorClient(nproc=config['cs_worker'],
                                    host=config['cs_server'],
                                    soldir=solpath,
                                    user=config['cs_user'],
                                    ssh_opts=config['cs_ssh_opts'],
                                    mp_debug=config['mp_debug'])
    else:
        raise ValueError("Unknown evaluator mode")
    evaluator.set_model(mfem_model)

    if name != 'Probe':  # probe evaluator does not load solfiles.
        evaluator.set_solfiles(solfiles)
        evaluator.load_solfiles()

    evaluator.make_agents(name, params, **kwargs)

    return evaluator


def area_tri(vertices):
    # area of triangles
    v = vertices
    p1 = v[:, 0, :]-v[:, 1, :]
    p2 = v[:, 0, :]-v[:, 2, :]
    return (np.sqrt(np.sum(np.cross(p1, p2)**2, 1)))/2.0



'''
   Global Named List

      A Named List having a capability to share its data among MPI proc.

      node 1: {1: [1,2,3], 2: [1,2]}
      node 2: {1: [4], 4: [5]}

      d = GNameList(dtype=<type>, gkey=<iter>, freezekey=<bool>)

      d.freezekey
         prohibit to add new key
      d.sharekeys:
         to make sure all node has the same dict keys.
         (note, in MPI communication, member access is done using self.gkey)
      d.allgather:
         {1: [1,2,3,4], 2:[1,2], 4:[5]}
      d.gather(root=<int>, distributed=<bool>):
         gather to particular node
         if distributed: n-th item is gather to n = myid % nprc 

'''
import numpy as np

from petram.mfem_config import use_parallel
if use_parallel:
    try:
        from mpi4py import MPI
        myid = MPI.COMM_WORLD.rank
        #nprc = MPI.COMM_WORLD.size
        #this is because even if nprc > 1, we may want this behave as if nprc = 1
        comm  = MPI.COMM_WORLD
        from mfem.common.mpi_debug import nicePrint, niceCall        
        from petram.helper.mpi_recipes import allgather, allgather_vector, gather_vector
        hasMPI = True
    except ImportError:
        hasMPI = False
else:
    hasMPI = False
    
class GlobalNamedList(dict):
    def __init__(self, *args, **kwargs):
        self.dtype = kwargs.pop('dtype', int)
        self._gkey = kwargs.pop('gkey', ())# global keys (ordered)
        self._freezekey = kwargs.pop('freezekey', False)
        super(GlobalNamedList, self).__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        if self._freezekey and key not in self:
            raise KeyError("{} is not a legal key of this StricDict".format(repr(key)))
        dict.__setitem__(self, key, value)

    def __getitem__(self, key):
        if not self._freezekey and not key in self:
            dict.__setitem__(self, key, [])
        return dict.__getitem__(self, key)

    @property
    def globalkeys(self):
        return self._gkey
    @property
    def freezekey(self):
        return self._freezekey
    @freezekey.setter
    def freezekey(self, value):
        self._freezekey = value

    def setlists(self, keys, values):
        '''
        [1,2,4,1], [1,2,3,7] -> [1:[1,7], 2:[2], 4:[1]]
        values are stored as numpy array

        '''
        values = np.array(values, copy=False, dtype=self.dtype)
        k, idx = np.unique(keys, return_inverse=True)
        for i in range(len(k)):
            dict.__setitem__(self, k[i], values[np.where(idx == i)[0]])
            
    def sort(self):
        '''
        sort eacy item
        '''
        for k in self:
            self[k] = np.sort(self[k]).astype(self.dtype, copy=False)
        return self
    
    def unique(self):            
        for k in self:
            self[k] = np.unique(self[k]).astype(self.dtype, copy=False)
        return self
            
    def sharekeys(self):
        keys = list(self.keys())
        if not hasMPI:
            keys = np.unique(keys)
            self._gkey = keys
            return self
        
        keys = sum(allgather(keys), [])
        keys = np.unique(keys)
        freezekey = self._freezekey
        for k in keys:
            if not k in self: self[k] = np.atleast_1d([]).astype(self.dtype)
        self._freezekey = freezekey
        self._gkey = keys
        return self
        
    def allgather(self, overwrite=True):
        if not hasMPI: return
        dest = self if overwrite else GlobalNamedList()       
        for key in self._gkey:
            if key in self:
                data = np.array(self[key], dtype=self.dtype, copy=False)
            else:
                data = np.atleast_1d([]).astype(self.dtype)
            data = allgather_vector(data)
            dest[key] = data
        return dest
                                
    def gather(self, nprc, root=None, distribute = False, overwrite=True):
        # we don't take nprc from MPI.COMM_WROLD (see a comment above)
        if nprc == 1:
            dest = self if overwrite else GlobalNamedList()            
            if not overwrite:
                for key in self:
                    dest[key] = np.array(self[key]).copy()
            return dest
        dest = self if overwrite else GlobalNamedList()
        for j, key in enumerate(self._gkey):
            if key in self:
                data = np.array(self[key], dtype=self.dtype, copy=False)
            else:
                data = np.atleast_1d([]).astype(self.dtype)
            r = 0 if root is None else root
            r = j % nprc if distribute else r
            data = gather_vector(data, root=r)
            if myid == r:
                dest[key] = data
            else:
                if key in dest: del dest[key]
        return dest
        
    def bcast(self, nprc, root = None, distributed = False, overwrite=True):
        # we don't take nprc from MPI.COMM_WROLD (see a comment above)
        if nprc == 1:
            dest = self if overwrite else GlobalNamedList()            
            if not overwrite:
                for key in self:
                    dest[key] = np.array(self[key]).copy()
            return dest
        dest = self if overwrite else GlobalNamedList()      
        for j, key in enumerate(self._gkey):
            r = 0 if root is None else root
            r = j % nprc if distributed else r
            
            if myid == r:
                if key in self:
                    data = np.array(self[key], dtype=self.dtype, copy=False)
                else:
                    data = np.atleast_1d([]).astype(self.dtype)
            else:
                data = None
            data = comm.bcast(data, root=r)
            dest[key] = data
        return dest



def attribute_set_names(name, suffix):
    '''
    ex. epsilonr -> expilonr_xx, _xy,,,,
    '''
    return [name + '_'+ x for x in suffix]

def init_multivalidator_attribute_set(d, name, suffix, values, use):
    names = attribute_set_names(name, suffix)
    for n, v in names, values:
        d[n] = v
    d['use_m_'+name] = use

class ValidatorInputMixIn(object):        
    def validator_elp_panel(self, row, col, name):
        elp = [[None, None, 0, {'validator': self.check_phys_expr,
                            'validator_param': name},]]
    
        return elp
    
    def get_alidator_panel1_value(self, name, suffix):
        '''
        get attribute xxx_m, xxx_suffix, use_m_xxx for
        elp panel
        '''         
        return [[str(getattr(self, name))],]
                
    def import_validator_panel1_value(self, v, name, suffix):
        '''
        set attribute xxx_m, xxx_suffix, use_m_xxx using
        elp panel value
        '''         
        setattr(self, name,  str(v[0]))

    def init_multivalidator_attribute_set(self, d, name, suffix, values, use):
        names = attribute_set_names(name, suffix)
        for n, v in names, values:
            d[n] = v
        d['use_m_'+name] = use
                
    
    def multivalicator_elp_panel(self, row, col, name, suffix):
        '''
        generate elp (row x col)
        '''
        a = [{'validator': self.check_phys_expr,
              'validator_param':name + '_' + s} for s in suffix]
                   
        elp1 = [[None, None, 43, {'row': row,
                              'col': col,
                             'text_setting': a}],]
        elp2 = [[None, None, 0, {'validator': self.check_phys_array_expr,
                             'validator_param': name+'_m'},]]
                   
        elp = [None, None, 34, ({'text': name + '   ',
                             'choices': ['Elemental Form', 'Array Form'],
                             'call_fit': False},
                             {'elp': elp1},  
                             {'elp': elp2},),],
        return elp
   
    def form_name(self, v):
        if v: return 'Array Form'
        else: return 'Elemental Form'
        
    def get_multivalidator_panel1_value(self, name, suffix):
        '''
        get attribute xxx_m, xxx_suffix, use_m_xxx for
        elp panel
        '''         
        return [self.form_name(getattr(self, 'use_m_' + name)), 
                [[str(getattr(self, name + s)) for s in suffix]],
                [str(getattr(self, name + '_m'))]]

    def import_multivalidator_panel1_value(self, v, name, suffix):
        '''
        set attribute xxx_m, xxx_suffix, use_m_xxx using
        elp panel value
        '''         
        setattr(self, 'use_m_'+name,  (str(v[0]) == 'Array Form'))
        for k, s in enumerate(suffix):
            setattr(self, name + '_' + s, str(v[1][0][k]))
        setattr(self, name + '_m',  str(v[2][0]))
            

    def make_multivaliator_f_name(self, basename, suffix):
        if getattr(self, 'use_m_'+basename):
            names = ['_m']
            eval_expr = self.eval_phys_array_expr
        else:
            names = ['_' + x for x in suffix]
            eval_expr = self.eval_phys_expr            
              
        f_name = []
        for n in names:
           var, f_name0 = eval_expr(getattr(self, basename+n), basename + n)
           if f_name0 is None:
               f_name.append(var)
           else:
               f_name.append(f_name0)
        
        return f_name



from petram.solver.strumpack_model import Strumpack
from petram.solver.mumps_model import MUMPSMFEMSolverModel
from petram.solver.krylov import KrylovModel, StationaryRefinementModel
import os
import warnings

import numpy as np

from petram.namespace_mixin import NS_mixin
from petram.solver.iterative_model import (Iterative,
                                           IterativeSolver)
from petram.solver.solver_model import (SolverBase,
                                        SolverInstance)

from petram.model import Model
from petram.solver.solver_model import Solver, SolverInstance
from petram.solver.std_solver_model import StdSolver

from petram.mfem_config import use_parallel
if use_parallel:
    from petram.helper.mpi_recipes import *
    import mfem.par as mfem
else:
    import mfem.ser as mfem

import petram.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('MGSolver')

rprint = debug.regular_print('MGSolver')


class FineLevel(NS_mixin):
    def attribute_set(self, v):
        super(FineLevel, self).attribute_set(v)
        v["grid_level"] = 1
        return v

    def get_info_str(self):
        return 'Lv:' + str(self.grid_level)

    def get_special_menu(self, evt):
        return [["Set level", self.onSetLevel, None, ], ]

    def onSetLevel(self, evt):
        import wx
        from ifigure.utils.edit_list import DialogEditList

        diag = evt.GetEventObject().GetTopLevelParent()

        list6 = [["New level", self.grid_level, 0], ]
        value = DialogEditList(list6,
                               modal=True,
                               style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER,
                               tip=None,
                               parent=diag,
                               title='Set level...')
        if not value[0]:
            return

        self.grid_level = int(value[1][0])


class CoarsestLvlSolver:
    grid_level = 0

    def __init__(self):
        self.is_preconditioner = True


class FinestLvlSolver:
    ...


class RefinedLevel(FineLevel, SolverBase):
    hide_ns_menu = True
    has_2nd_panel = True

    def __init__(self, *args, **kwags):
        SolverBase.__init__(self, *args, **kwags)
        FineLevel.__init__(self, *args, **kwags)
        self.smoother_count = None

    def attribute_set(self, v):
        v = FineLevel.attribute_set(self, v)
        v = SolverBase.attribute_set(self, v)
        v["level_inc"] = "1"
        v["refinement_type"] = "P(order)"
        v["extra_constraints1"] = ""
        v["extra_constraints2"] = ""
        v["select_extra_constreint"] = False
        v["unselect_extra_constreint"] = False
        v["use_domain_refinement"] = False
        v["presmoother_count"] = "default"
        v["postsmoother_count"] = "default"
        v["cycle_max"] = "default"

        return v

    def panel1_param(self):
        panels = super(RefinedLevel, self).panel1_param()

        mm = [["constraints", "", 0, {}], ]

        p2 = [None, (self.select_extra_constreint, (self.extra_constraints1,)),
              27, ({"text": "Enable contributions"}, {"elp": mm},)]
        p3 = [None, (self.unselect_extra_constreint, (self.extra_constraints2,)),
              27, ({"text": "Disable contributions"}, {"elp": mm},)]
        p4 = [["pre-smoother #count", "", 0, {}],
              ["post-smoother #count", "", 0, {}],
              ["cycle max (V:1, W;2)", "", 0, {}], ]

        panels.extend([["refinement type", self.refinement_type, 1,
                        {"values": ["P(order)", "H(mesh)"]}],
                       ["level increment", "", 0, {}],
                       p2, p3])
        panels.extend(p4)
        return panels

    def get_panel1_value(self):
        value = list(super(RefinedLevel, self).get_panel1_value())
        value.append(self.refinement_type)
        value.append(self.level_inc)
        v2 = (self.select_extra_constreint, (self.extra_constraints1,))
        v3 = (self.unselect_extra_constreint, (self.extra_constraints2,))
        v4 = [self.presmoother_count,
              self.postsmoother_count,
              self.cycle_max]

        value.append(v2)
        value.append(v3)
        value.extend(v4)

        return value

    def import_panel1_value(self, v):
        super(RefinedLevel, self).import_panel1_value(v[:-7])
        self.refinement_type = v[-7]
        self.level_inc = v[-6]
        self.select_extra_constreint = v[-5][0]
        self.extra_constraints1 = v[-5][1][0]
        self.unselect_extra_constreint = v[-4][0]
        self.extra_constraints2 = v[-4][1][0]
        self.presmoother_count = v[-3]
        self.postsmoother_count = v[-2]
        self.cycle_max = v[-1]

    def panel2_param(self):
        mm = [["constraints", "", 0, {}], ]
        p2 = [None, (self.use_domain_refinement, (self._sel_index,)),
              27, ({"text": "Resrict domain (H-only)"}, {"elp": mm},)]

        return [p2, ]

    def get_panel2_value(self):
        txt = ",".join([str(x) for x in self._sel_index])
        v1 = (self.use_domain_refinement, (txt,))
        return (v1,)

    def import_panel2_value(self, v):
        self.use_domain_refinement = v[0][0]
        self._sel_index = [x for x in v[0][1][0].split(',')]

    def get_possible_child(self):
        from petram.solver.krylov import KrylovSmoother
        from petram.solver.block_smoother import DiagonalSmoother
        return [KrylovSmoother, DiagonalSmoother]

    def get_phys(self):
        my_solve_step = self.get_solve_root()
        return my_solve_step.get_phys()

    def get_phys_range(self):
        my_solve_step = self.get_solve_root()
        return my_solve_step.get_phys_range()

    def prepare_solver(self, opr, engine):
        if self.smoother_count[1] == 0:
            for x in self.iter_enabled():
                return x.prepare_solver(opr, engine)
        else:
            for x in self.iter_enabled():
                return x.prepare_solver_with_multtranspose(opr, engine)
            # return x.prepare_solver(opr, engine)

    def adjust_physics(self, phys):
        names = [x.strip() for x in self.extra_constraints1.split(',')]
        if self.select_extra_constreint:
            for o in phys.walk():
                fname = o.fullname()
                for n in names:
                    if fname.find(n) != -1:
                        o.enabled = True
                        dprint1("!!!!! enabling", fname)

        names = [x.strip() for x in self.extra_constraints2.split(',')]
        if self.unselect_extra_constreint:
            for o in phys.walk():
                fname = o.fullname()
                for n in names:
                    if fname.find(n) != -1:
                        o.enabled = False
                        dprint1("!!!!!disabling", fname)

    @classmethod
    def fancy_tree_name(self):
        return 'Refined'

    @property
    def is_iterative(self):
        return True


class CoarseIterative(KrylovModel, CoarsestLvlSolver):

    def __init__(self, *args, **kwargs):
        KrylovModel.__init__(self, *args, **kwargs)
        CoarsestLvlSolver.__init__(self)

    @classmethod
    def fancy_menu_name(self):
        return 'Kryrov'

    @classmethod
    def fancy_tree_name(self):
        return 'Kryrov'

    def get_info_str(self):
        return 'Coarsest:Lv0'

    def prepare_solver(self, opr, engine):
        solver = self.do_prepare_solver(opr, engine)

        if self.is_preconditioner:
            solver.iterative_mode = False
        else:
            solver.iterative_mode = True

        return solver


class CoarseRefinement(StationaryRefinementModel, CoarsestLvlSolver):
    def __init__(self, *args, **kwargs):
        StationaryRefinementModel.__init__(self, *args, **kwargs)
        CoarsestLvlSolver.__init__(self)

    @classmethod
    def fancy_menu_name(self):
        return 'Refinement'

    @classmethod
    def fancy_tree_name(self):
        return 'Refinement'

    def get_info_str(self):
        return 'Coarsest:Lv0'


class CoarseMUMPS(MUMPSMFEMSolverModel, CoarsestLvlSolver):
    def __init__(self, *args, **kwargs):
        MUMPSMFEMSolverModel.__init__(self, *args, **kwargs)
        CoarsestLvlSolver.__init__(self)

    @classmethod
    def fancy_menu_name(self):
        return 'MUMPS'

    @classmethod
    def fancy_tree_name(self):
        return 'Direct'

    def get_info_str(self):
        return 'MUMPS:Lv0'


class CoarseStrumpack(Strumpack, CoarsestLvlSolver):
    @classmethod
    def fancy_menu_name(self):
        return 'STRUMPACK'

    @classmethod
    def fancy_tree_name(self):
        return 'Direct'

    def get_info_str(self):
        return 'STRUMPACK:Lv0'


class FineIterative(KrylovModel, FinestLvlSolver):
    @classmethod
    def fancy_menu_name(self):
        return 'Kryrov'

    @classmethod
    def fancy_tree_name(self):
        return 'Iterative'

    def get_info_str(self):
        return 'Finest'

    def get_possible_child(self):
        return []

    def get_possible_child_menu(self):
        return []

    def prepare_solver(self, opr, engine):
        solver = self.do_prepare_solver(opr, engine)
        solver.iterative_mode = True

        return solver


class FineRefinement(StationaryRefinementModel, FinestLvlSolver):
    @classmethod
    def fancy_menu_name(self):
        return 'StationaryRefinement'

    @classmethod
    def fancy_tree_name(self):
        return 'Refinement'

    def get_info_str(self):
        return 'Finest'

    def get_possible_child(self):
        return []

    def get_possible_child_menu(self):
        return []

    def prepare_solver(self, opr, engine):
        solver = self.do_prepare_solver(opr, engine)
        solver.iterative_mode = True

        return solver


class MultiLvlStationarySolver(StdSolver):
    @classmethod
    def fancy_menu_name(self):
        return 'Stationary(MultiLevel)'

    @classmethod
    def fancy_tree_name(self):
        return 'Stationary'

    def attribute_set(self, v):
        super(MultiLvlSolver, self).attribute_set(v)
        v['merge_real_imag'] = True
        v['use_block_symmetric'] = False
        v["presmoother_count"] = "1"
        v["postsmoother_count"] = "1"
        v["cycle_max"] = "2"
        v['assemble_real'] = True
        v["debug_cycle_level"] = False
        v["debug_residuals"] = False

        return v

    def panel1_param(self):
        panels = super(MultiLvlStationarySolver, self).panel1_param()

        mm = [[None, self.use_block_symmetric, 3,
               {"text": "block symmetric format"}], ]

        p2 = [[None, (self.merge_real_imag, (self.use_block_symmetric,)),
               27, ({"text": "Use ComplexOperator"}, {"elp": mm},)], ]
        panels.extend(p2)

        p3 = [["pre-smoother #count", "", 0, {}],
              ["post-smoother #count", "", 0, {}],
              ["cycle max (V:1, W;2)", "", 0, {}], ]
        panels.extend(p3)

        return panels

    def get_panel1_value(self):
        value = list(super(MultiLvlStationarySolver, self).get_panel1_value())
        value.append((self.merge_real_imag, [self.use_block_symmetric, ]))
        value.append(self.presmoother_count)
        value.append(self.postsmoother_count)
        value.append(self.cycle_max)
        return value

    def import_panel1_value(self, v):
        super(MultiLvlStationarySolver, self).import_panel1_value(v[:-4])
        self.merge_real_imag = bool(v[-4][0])
        self.use_block_symmetric = bool(v[-4][1][0])
        self.presmoother_count = v[-3]
        self.postsmoother_count = v[-2]
        self.cycle_max = v[-1]

    def allocate_solver_instance(self, engine):
        if self.clear_wdir:
            engine.remove_solfiles()

        instance = MLInstance(self, engine)
        return instance

    def get_matrix_weight(self, timestep_config):  # , timestep_weight):
        if timestep_config[0]:
            return [1, 0, 0]
        else:
            return [0, 0, 0]

    def does_solver_choose_linearsystem_type(self):
        return True

    def get_linearsystem_type_from_solvermodel(self):
        assemble_real = self.assemble_real
        phys_real = self.get_solve_root().is_allphys_real()

        if phys_real:
            if assemble_real:
                dprint1("Use assemble-real is only for complex value problem !!!!")
                return 'blk_interleave'
            else:
                return 'blk_interleave'

        # below phys is complex

        # merge_real_imag -> complex operator
        if self.merge_real_imag and self.use_block_symmetric:
            return 'blk_merged_s'
        elif self.merge_real_imag and not self.use_block_symmetric:
            return 'blk_merged'
        elif assemble_real:
            return 'blk_interleave'
        else:
            assert False, "complex problem must assembled using complex operator or expand as real value problem"
        # return None

    def verify_setting(self):
        '''
        has to have one coarse solver
        '''
        isvalid = True
        txt = ''
        txt_long = ''
        cs = self.get_coarsest_solvers()
        if len(cs) != 1:
            isvalid = False
            txt = 'Invlid MultiLvlSolver configuration'
            txt_long = 'Number of active coarse level solver must be one.'

        levels, klevels, setting = self._get_level_solvers()

        if list(range(len(levels))) != klevels:
            isvalid = False
            txt = 'Invlid MultiLvlSolver configuration'
            txt_long = 'Grid levels are not set properly'

        finest = self.get_active_solver(cls=FinestLvlSolver)
        if finest is not None and len(levels) == 1:
            isvalid = False
            txt = 'Invlid MultiLvlSolver configuration'
            txt_long = 'There is no multi-level setting, while finest solver is set'

        return isvalid, txt, txt_long

    def get_possible_child(self):
        return (CoarseMUMPS,
                CoarseStrumpack,
                CoarseRefinement,
                CoarseIterative,
                RefinedLevel,
                FineIterative,
                FineRefinement)

    def get_possible_child_menu(self):
        choice = [("CoarseSolver", CoarseMUMPS),
                  ("", CoarseStrumpack),
                  ("", CoarseIterative),
                  ("!", CoarseRefinement),
                  ("", RefinedLevel),
                  ("FineSolver", FineIterative),
                  ("!", FineRefinement), ]
        return choice

    def get_coarsest_solvers(self):
        s = []
        for x in self:
            child = self[x]
            if not child.is_enabled():
                continue
            if isinstance(child, CoarsestLvlSolver):
                s.append(child)
        return s

    def _get_level_solvers(self):
        levels = [self.get_coarsest_solvers()[0]]
        klevels = [0]

        level_setting = {}

        for x in self:
            child = self[x]
            if not child.is_enabled():
                continue
            if isinstance(child, CoarsestLvlSolver):
                continue
            if isinstance(child, RefinedLevel):
                levels.append(child)
                klevels.append(child.grid_level)

                presmoother_count = int(self.presmoother_count
                                        if child.presmoother_count.lower() == 'default'
                                        else child.presmooth_count)
                postsmoother_count = int(self.postsmoother_count
                                         if child.postsmoother_count.lower() == 'default'
                                         else child.presmooth_count)
                cycle_max = int(self.cycle_max
                                if child.cycle_max.lower() == 'default'
                                else child.cycle_max)

                level_setting[child.grid_level] = (
                    presmoother_count, postsmoother_count, cycle_max)

        idx = np.argsort(klevels)
        levels = [levels[i] for i in idx]
        klevels = [int(klevels[i]) for i in idx]
        return levels, klevels, level_setting

    def get_level_solvers(self):
        return self._get_level_solvers()[0]

    def get_level_setting(self):
        return self._get_level_solvers()[-1]

    def create_refined_levels(self, engine, lvl):
        '''
        lvl : refined level number (1, 2, 3, ....)
              1 means "AFTER" 1 refinement
        '''
        levels = self.get_level_solvers()
        for l in levels:
            l.smoother_count = (int(self.presmoother_count),
                                int(self.postsmoother_count))

        if lvl >= len(levels):
            return False

        level = levels[lvl]

        target_phys = self.get_target_phys()
        refine_type = level.refinement_type[0]  # P or H
        refine_inc = int(level.level_inc)
        if level.use_domain_refinement:
            refine_dom = [int(x) for x in level._sel_index]
        else:
            refine_dom = None

        for phys in target_phys:
            dprint1("Adding refined level for " + phys.name())
            level.adjust_physics(phys)
            engine.prepare_refined_level(phys, refine_type,
                                         inc=refine_inc,
                                         refine_dom=refine_dom)

        engine.level_idx = lvl
        for phys in target_phys:
            engine.get_true_v_sizes(phys)

        return True

    @debug.use_profiler
    def run(self, engine, is_first=True, return_instance=False):
        dprint1("Entering run (is_first=", is_first, ")", self.fullpath())
        if self.clear_wdir:
            engine.remove_solfiles()

        instance = self.allocate_solver_instance(engine)
        instance.set_blk_mask()
        if return_instance:
            return instance

        instance.configure_probes(self.probe)

        if self.init_only:
            engine.sol = engine.assembled_blocks[1][0]
            instance.sol = engine.sol
        else:
            if is_first:
                instance.assemble()
                is_first = False
            instance.solve()

        instance.save_solution(ksol=0,
                               skip_mesh=False,
                               mesh_only=False,
                               save_parmesh=self.save_parmesh)
        engine.sol = instance.sol

        instance.save_probe()

        dprint1(debug.format_memory_usage())
        return is_first


MultiLvlSolver = MultiLvlStationarySolver


class MLInstance(SolverInstance):
    def __init__(self, gui, engine):
        SolverInstance.__init__(self, gui, engine)
        self.assembled = False
        self.linearsolver = None

    @property
    def blocks(self):
        return self.engine.assembled_blocks

    def compute_A(self, M, B, X, mask_M, mask_B):
        '''
        M[0] x = B

        return A and isAnew
        '''
        return M[0], np.any(mask_M[0])

    def compute_rhs(self, M, B, X):
        '''
        M[0] x = B
        '''
        return B

    def do_assemble(self, inplace=True):
        engine = self.engine
        phys_target = self.get_phys()
        phys_range = self.get_phys_range()

        # use get_phys to apply essential to all phys in solvestep
        dprint1("Asembling system matrix",
                [x.name() for x in phys_target],
                [x.name() for x in phys_range])
        engine.run_verify_setting(phys_target, self.gui)
        engine.run_assemble_mat(phys_target, phys_range)
        engine.run_assemble_b(phys_target)
        engine.run_fill_X_block()

        self.engine.run_assemble_blocks(self.compute_A,
                                        self.compute_rhs,
                                        inplace=inplace)
        #A, X, RHS, Ae, B, M, names = blocks

    def assemble(self, inplace=True):
        engine = self.engine

        levels = self.gui.get_level_solvers()

        enabled_flag = engine.model.gather_enebled_flags(engine.model['Phys'])

        for l, lvl in enumerate(levels):
            engine.level_idx = l

            if isinstance(lvl, RefinedLevel):
                for phys in self.get_phys():
                    lvl.adjust_physics(phys)
                engine.assign_sel_index()

            self.do_assemble(inplace)

        self.assembled = True

        engine.model.apply_enebled_flags(engine.model['Phys'], enabled_flag)

    @property
    def is_iterative(self):
        levels = self.gui.get_level_solvers()
        check = [x.is_iterative for x in levels]
        return np.any(check)

    def finalize_linear_system(self, level):
        engine = self.engine
        engine.level_idx = level
        ls_type = self.ls_type

        A, X, RHS, Ae, B, M, depvars = self.blocks

        mask = self.blk_mask
        engine.copy_block_mask(mask)

        depvars = [x for i, x in enumerate(depvars) if mask[0][i]]

        AA = engine.finalize_matrix(A, mask, not self.phys_real,
                                    format=self.ls_type)
        BB = engine.finalize_rhs([RHS], A, X[0], mask, not self.phys_real,
                                 format=ls_type)

        XX = engine.finalize_x(X[0], RHS, mask, not self.phys_real,
                               format=ls_type)
        return BB, XX, AA

    def finalize_linear_systems(self):
        '''
        call finalize_linear_system for all levels
        '''
        levels = self.gui.get_level_solvers()

        finalized_ls = []
        for k, _lvl in enumerate(levels):
            BB, XX, AA = self.finalize_linear_system(k)
            finalized_ls.append((BB, XX, AA))

        self.finalized_ls = finalized_ls

    def create_solvers(self):
        engine = self.engine
        levels = self.gui.get_level_solvers()
        finest = self.gui.get_active_solver(cls=FinestLvlSolver)

        if finest is None and len(levels) == 1:
            levels[0].is_preconditioner = False

        solvers = []
        for lvl, solver_model in enumerate(levels):
            engine.level_idx = lvl
            opr = self.finalized_ls[lvl][-1]
            s = solver_model.prepare_solver(opr, engine)
            solvers.append(s)

        if finest is not None:
            opr = self.finalized_ls[lvl][-1]
            finestsolver = finest.prepare_solver(opr, engine)
        else:
            finestsolver = None
        return solvers, finestsolver

    def create_prolongations(self):
        engine = self.engine
        levels = self.gui.get_level_solvers()

        prolongations = []
        for k, _lvl in enumerate(levels):
            if k == len(levels)-1:
                break
            _BB, XX, AA = self.finalized_ls[k]
            P = fill_prolongation_operator(
                engine, k, XX, AA, self.ls_type, self.phys_real)
            prolongations.append(P)

        return prolongations

    def assemble_rhs(self):
        assert False, "assemble_rhs is not implemented"

    def _solve(self, update_operator=True):
        engine = self.engine

        self.finalize_linear_systems()

        smoothers, finestsolver = self.create_solvers()
        prolongations = self.create_prolongations()

        operators = [x[-1] for x in self.finalized_ls]

        #solall = linearsolver.Mult(BB, case_base=0)
        if len(smoothers) > 1:
            mg = generate_MG(operators, smoothers, prolongations,
                             presmoother_count=int(self.gui.presmoother_count),
                             postsmoother_count=int(self.gui.postsmoother_count))

        else:
            mg = None

        BB, XX, AA = self.finalized_ls[-1]

        if finestsolver is not None:
            finestsolver.SetPreconditioner(mg)
            finestsolver.Mult(BB[0], XX)
            solall = np.transpose(np.vstack([XX.GetDataArray()]))
        elif mg is not None:
            mg.Mult(BB[0], XX)
            solall = np.transpose(np.vstack([XX.GetDataArray()]))
        else:
            # this makes sense if coarsest smoother is direct solver
            smoothers[0].Mult(BB[0], XX)
            solall = np.transpose(np.vstack([XX.GetDataArray()]))

        if not self.phys_real:
            from petram.solver.solver_model import convert_realblocks_to_complex
            merge_real_imag = self.ls_type in ["blk_merged", "blk_merged_s"]
            solall = convert_realblocks_to_complex(solall, AA, merge_real_imag)

        engine.level_idx = len(self.finalized_ls)-1
        A = engine.assembled_blocks[0]
        X = engine.assembled_blocks[1]
        A.reformat_distributed_mat(solall, 0, X[0], self.blk_mask)

        self.sol = X[0]

        # store probe signal (use t=0.0 in std_solver)
        for p in self.probe:
            p.append_sol(X[0])

        return True

    def solve(self, update_operator=True):
        '''
        test version calls one V cycle written in Python
        '''
        engine = self.engine

        self.finalize_linear_systems()

        smoothers, finestsolver = self.create_solvers()
        prolongations = self.create_prolongations()

        operators = [x[-1] for x in self.finalized_ls]

        #offsets0 = operators[0].RowOffsets().ToList()
        #offsets1 = operators[1].RowOffsets().ToList()

        if len(smoothers) > 1:
            lvl = engine.level_idx
            is_complex = not self.phys_real

            ess_tdofs_arr = []
            for lvl in range(len(operators)):
                engine.level_idx = lvl
                esstdofs0 = engine.collect_local_ess_TDofs(
                    operators[lvl], self.ls_type, is_complex)
                ess_tdofs_arr.append(esstdofs0)

            #solall = linearsolver.Mult(BB, case_base=0)
            level_setting = self.gui.get_level_setting()

            presmoother_count = {
                key: level_setting[key][0] for key in level_setting}
            postsmoother_count = {
                key: level_setting[key][1] for key in level_setting}
            cycle_max = {key: level_setting[key][2] for key in level_setting}

            mg = PyMG(operators,
                      smoothers,
                      prolongations,
                      ess_tdofs=ess_tdofs_arr,
                      presmoother_count=presmoother_count,
                      postsmoother_count=postsmoother_count,
                      cycle_max=cycle_max,
                      debug1=self.gui.debug_cycle_level,
                      debug2=self.gui.debug_residuals)

        else:
            mg = None

        BB, XX, AA = self.finalized_ls[-1]

        if finestsolver is not None:
            finestsolver.SetPreconditioner(mg)
            finestsolver.Mult(BB[0], XX)
            solall = np.transpose(np.vstack([XX.GetDataArray()]))
        elif mg is not None:
            mg.Mult(BB[0], XX)
            solall = np.transpose(np.vstack([XX.GetDataArray()]))
        else:
            # this makes sense if coarsest smoother is direct solver
            smoothers[0].Mult(BB[0], XX)
            solall = np.transpose(np.vstack([XX.GetDataArray()]))

        solall = np.transpose(np.vstack([XX.GetDataArray()]))

        if not self.phys_real:
            from petram.solver.solver_model import convert_realblocks_to_complex
            merge_real_imag = self.ls_type in ["blk_merged", "blk_merged_s"]
            solall = convert_realblocks_to_complex(solall, AA, merge_real_imag)

        engine.level_idx = len(self.finalized_ls)-1
        A = engine.assembled_blocks[0]
        X = engine.assembled_blocks[1]
        A.reformat_distributed_mat(solall, 0, X[0], self.blk_mask)

        self.sol = X[0]

        # store probe signal (use t=0.0 in std_solver)
        for p in self.probe:
            p.append_sol(X[0])

        return True


def generate_MG(operators, smoothers, prolongations,
                presmoother_count=1,
                postsmoother_count=1):

    own_operators = [False]*len(operators)
    own_smoothers = [False]*len(smoothers)
    own_prolongations = mfem.boolArray([False]*len(prolongations))

    ops = mfem.OperatorPtrArray(operators)
    # mg = mfem.Multigrid(operators, smoothers, prolongations,
    mg = mfem.Multigrid(ops, smoothers, prolongations,
                        own_operators, own_smoothers, own_prolongations)

    mg.SetCycleType(mfem.Multigrid.CycleType_VCYCLE,
                    presmoother_count, postsmoother_count)

    return mg


class PyMG(mfem.PyIterativeSolver):
    def __init__(self, operators, smoothers, prolongations,
                 ess_tdofs=None, presmoother_count=1, postsmoother_count=1,
                 debug1=True, debug2=True, cycle_max=1):
        self.operators = operators
        self.smoothers = smoothers
        self.prolongations = prolongations
        self.presmoother_count = presmoother_count
        self.postsmoother_count = postsmoother_count
        self.ess_tdofs = ess_tdofs

        self.debug1 = debug1
        self.debug2 = debug2

        self.cycle_rel_tol = 0.01
        self.cycle_max = cycle_max

        if use_parallel:
            from mpi4py import MPI
            args = (MPI.COMM_WORLD,)
        else:
            args = tuple()
        mfem.PyIterativeSolver.__init__(self, *args)

    def Mult(self, x, y):
        '''
        call cycle_max times of perform_one_cycle
        '''
        y.Assign(0.0)

        lvl = len(self.ess_tdofs)-1

        correction = mfem.Vector(y.Size())
        correction.Assign(0.0)
        err = mfem.Vector(x.Size())
        err.Assign(x)
        tmp = mfem.Vector(x.Size())

        for i in range(self.cycle_max[lvl]):
            self.perform_one_cycle(err, correction)

            self.operators[lvl].Mult(correction, tmp)
            tmp *= -1
            tmp += err
            err.Assign(tmp)

            y += correction

    def perform_one_cycle(self, x, y, lvl=None):

        if lvl is None:
            # start from finest level
            lvl = len(self.ess_tdofs)-1

        if self.debug1:
            dprint1("========\n")
            dprint1("Entering Cycle lvl =  ", lvl)
        if lvl == 0:
            if self.debug2:
                dprint1("    - error on essential at level = 0",
                        np.sum(np.abs(x.GetDataArray()[self.ess_tdofs[0]])))
                dprint1("    - NormInf before level0 solve", x.Normlinf())

            y.Assign(0.0)
            self.smoothers[0].Mult(x, y)

            if self.debug2:
                dprint1("    - NormInf after level0 solve", y.Normlinf())
                dprint1("    - correction on essential at coarse level",
                        np.sum(np.abs(y.GetDataArray()[self.ess_tdofs[0]])))

                tmp = mfem.Vector(y.Size())
                self.operators[0].Mult(y, tmp)
                tmp -= x
                dprint1("    - level0 linear inverse error (L2): ", tmp.Norml2())

            if self.debug1:
                dprint1("Exiting Cycle lvl =  ", lvl)
                dprint1("========\n")
            return

        if self.debug2:
            dprint1("")
            dprint1("  - residual on essential at the start of level",
                    np.sum(np.abs(x.GetDataArray()[self.ess_tdofs[1]])))
            dprint1("  - initial residual L2", x.Norml2())

        err = mfem.Vector(x.Size())
        err.Assign(x)

        y0 = mfem.Vector(x.Size())
        y.Assign(0.0)

        for jjj in range(self.presmoother_count[lvl]):
            if self.debug1:
                txt = str(jjj) + '/' + str(self.presmoother_count[lvl])
                dprint1("   -  Performing pre-smooth " +
                        txt + " : level = " + str(lvl))

            self.smoothers[lvl].iterative_mode = False
            y0.Assign(0.0)

            if self.debug2:
                dprint1("    - resdidual on essential before presmooth",
                        np.sum(np.abs(err.GetDataArray()[self.ess_tdofs[lvl]])))

            self.smoothers[lvl].Mult(err, y0)

            y += y0
            self.operators[lvl].Mult(y, err)
            err *= -1
            err += x

            if self.debug2:
                dprint1("    - residual on essential after pre-smooth",
                        np.sum(np.abs(err.GetDataArray()[self.ess_tdofs[lvl]])))

        if self.debug2:
            dprint1(
                "  correction (L2) before adding prolonged correction: ", y.Norml2())

        # prepare err passed to lower level
        lvl2 = lvl - 1
        lvl2_width = self.operators[lvl2].Width()
        err2 = mfem.Vector(lvl2_width)
        self.prolongations[lvl2].MultTranspose(err, err2)
        y2 = mfem.Vector(lvl2_width)

        # (zeroing the error sent to the lower level)   <--- this works
        err2.GetDataArray()[self.ess_tdofs[lvl2]] = 0.0

        if self.debug2:
            dprint1("    - error on essential given to a coarse level",
                    np.sum(np.abs(err2.GetDataArray()[self.ess_tdofs[lvl2]])))

        # calling lower levels
        #   cycle max = 1 (V-cycle)
        #   cycle max = 2 (W-cycle)

        x2 = mfem.Vector(lvl2_width)
        x2.Assign(0.0)
        err2_L2 = err2.Norml2()
        if use_parallel:
            err2_L2 = np.sqrt(
                np.sum([x**2 for x in MPI.COMM_WORLD.allgather(err2_L2)]))
        rel_improve0 = 1.0
        for i in range(self.cycle_max[lvl]):
            self.perform_one_cycle(err2, y2, lvl=lvl2)
            x2 += y2
            tmp = mfem.Vector(x2.Size())
            self.operators[lvl2].Mult(y2, tmp)
            err2 -= tmp
            err2norm = err2.Norml2()
            if use_parallel:
                err2norm = np.sqrt(
                    np.sum([x**2 for x in MPI.COMM_WORLD.allgather(err2norm)]))

            rel_improve = err2norm/err2_L2

            if self.debug2:
                dprint1(str(i)+" th cycle checking cycle error lvl = :" + str(lvl))
                dprint1("correction L2/ rel_improve", y2.Norml2(), rel_improve)
                dprint1("change of improvement", np.abs(
                    np.abs(rel_improve0/rel_improve)-1))

            if rel_improve < self.cycle_rel_tol:
                break
            if np.abs(np.abs(rel_improve0/rel_improve)-1) < 0.1:
                break
            rel_improve0 = rel_improve

        y2 = x2

        # (zeroing the correction given from the lower level)   <--- this works
        y2.GetDataArray()[self.ess_tdofs[lvl2]] = 0.0

        self.prolongations[lvl2].Mult(y2, y0)
        # (zeroing after MUMPS after prolongation)
        # tmp.GetDataArray()[self.ess_tdofs[1]] = 0.0  # <--- does not works
        y += y0

        if self.debug2:
            dprint1("    - correction on essential prolonged from lower level",
                    np.sum(np.abs(y0.GetDataArray()[self.ess_tdofs[lvl]])))
            dprint1(
                "    - correction (L2) after adding prolonged correction: ", y.Norml2())

        for jjj in range(self.postsmoother_count[lvl]):
            if self.debug1:
                txt = str(jjj) + '/' + str(self.postsmoother_count[lvl])
                dprint1("   -  Performing post-smooth " +
                        txt + " : level = " + str(lvl))

            # compute error
            self.operators[lvl].Mult(y, err)
            err *= -1
            err += x

            if self.debug2:
                dprint1("    - residual on essential before postsmooth",
                        np.sum(np.abs(err.GetDataArray()[self.ess_tdofs[lvl]])))
                dprint1("    - residual L2 before before postsmooth", err.Norml2())

            y0.Assign(0.0)
            self.smoothers[lvl].Mult(err, y0)
            # (zeroing the essentials of final correction)
            # y0.GetDataArray()[self.ess_tdofs[1]] = 0.0   <--- does not works
            y += y0

            if self.debug2:
                dprint1("    - correction on essential",
                        np.sum(np.abs(y0.GetDataArray()[self.ess_tdofs[lvl]])))
                dprint1("    - correction norm (L2) after postsmooth", y0.Norml2())

        if self.debug2:
            self.operators[lvl].Mult(y, err)
            err *= -1
            err += x

            dprint1("  - final correction on essential",
                    np.sum(np.abs(y.GetDataArray()[self.ess_tdofs[lvl]])))
            dprint1("  - finial norm of correction (L2): ", y.Norml2())
            dprint1("  - final residual L2", err.Norml2())
        if self.debug1:
            dprint1("Exiting Cycle lvl =  ", lvl)
            dprint1("========\n")


def fill_prolongation_operator(engine, level, XX, AA, ls_type, phys_real):
    engine.access_idx = 0
    P = None
    diags = []

    use_complex_opr = ls_type in ['blk_merged', 'blk_merged_s']

    widths = [XX.BlockSize(i) for i in range(XX.NumBlocks())]

    if not phys_real:
        if use_complex_opr:
            widths = [x//2 for x in widths]
        else:
            widths = [widths[i*2] for i in range(len(widths)//2)]

    cols = [0]
    rows = [0]

    for dep_var in engine.r_dep_vars:
        offset = engine.r_dep_var_offset(dep_var)

        tmp_cols = []
        tmp_rows = []
        tmp_diags = []

        # if use_complex_opr:
        #    mat = AA._linked_op[(offset, offset)]
        #    conv = mat.GetConvention()
        #    conv = (1 if mfem.ComplexOperator.HERMITIAN else -1)
        # else:
        #    conv = 1

        if engine.r_isFESvar(dep_var):
            h = engine.fespaces.get_hierarchy(dep_var)
            P = h.GetProlongationAtLevel(level)
            tmp_cols.append(P.Width())
            tmp_rows.append(P.Height())
            tmp_diags.append(P)
            if not phys_real:
                tmp_cols.append(P.Width())
                tmp_rows.append(P.Height())
                # if conv == -1:
                #    oo2 = mfem.ScaledOperator(P, -1)
                #    oo2._opr = P
                #    tmp_diags.append(oo2)
                # else:
                tmp_diags.append(P)
        else:
            tmp_cols.append(widths[offset])
            tmp_rows.append(widths[offset])
            tmp_diags.append(mfem.IdentityOperator(widths[offset]))

            if not phys_real:
                tmp_cols.append(widths[offset])
                tmp_rows.append(widths[offset])
                oo = mfem.IdentityOperator(widths[offset])
                # if conv == -1:
                #    oo2 = mfem.ScaledOperator(oo, -1)
                #    oo2._opr = oo
                #    tmp_diags.append(oo2)
                # else:
                tmp_diags.append(oo)

        if use_complex_opr:
            tmp_cols = [0] + tmp_cols
            tmp_rows = [0] + tmp_rows
            offset_c = mfem.intArray(tmp_cols)
            offset_r = mfem.intArray(tmp_rows)
            offset_c.PartialSum()
            offset_r.PartialSum()
            smoother = mfem.BlockOperator(offset_r, offset_c)
            smoother.SetDiagonalBlock(0, tmp_diags[0])
            smoother.SetDiagonalBlock(1, tmp_diags[1])
            smoother._smoother = tmp_diags
            cols.append(tmp_cols[1]*2)
            rows.append(tmp_rows[1]*2)
            diags.append(smoother)
        else:
            cols.extend(tmp_cols)
            rows.extend(tmp_rows)
            diags.extend(tmp_diags)

    ro = mfem.intArray(rows)
    co = mfem.intArray(cols)
    ro.PartialSum()
    co.PartialSum()

    P = mfem.BlockOperator(ro, co)
    for i, d in enumerate(diags):
        P.SetBlock(i, i, d)
    P._diags = diags
    return P


def generate_smoother(engine, level, blk_opr):
    from petram.engine import ParallelEngine
    assert not isinstance(
        engine, ParallelEngine), "Parallel is not supported"

    engine.access_idx = 0
    P = None
    diags = []
    cols = [0]
    ess_tdofs = []
    A, _X, _RHS, _Ae,  _B,  _M, _dep_vars = engine.assembled_blocks
    widths = A.get_local_col_widths()

    use_complex_opr = A.complex and (A.shape[0] == blk_opr.NumRowBlocks())

    for dep_var in engine.r_dep_vars:
        offset = engine.r_dep_var_offset(dep_var)

        tmp_cols = []
        tmp_diags = []
        if engine.r_isFESvar(dep_var):
            ess_tdof = mfem.intArray(engine.ess_tdofs[dep_var])
            ess_tdofs.append(ess_tdofs)
            opr = A[offset, offset]

            if use_complex_opr:
                mat = blk_opr._linked_op[(offset, offset)]
                conv = mat.GetConvention()
                conv == 1 if mfem.ComplexOperator.HERMITIAN else -1
                mat1 = mat._real_operator
                mat2 = mat._imag_operator
            else:
                conv = 1
                if A.complex:
                    mat1 = blk_opr.GetBlock(offset*2, offset*2)
                    mat2 = blk_opr.GetBlock(offset*2 + 1, offset*2 + 1)
                else:
                    mat1 = blk_opr.GetBlock(offset, offset)

            if A.complex:
                dd = opr.diagonal()
                diag = mfem.Vector(list(dd.real))
                rsmoother = mfem.OperatorChebyshevSmoother(mat1,
                                                           diag,
                                                           ess_tdof,
                                                           2)
                dd = opr.diagonal()*conv
                diag = mfem.Vector(list(dd.real))
                ismoother = mfem.OperatorChebyshevSmoother(mat1,
                                                           diag,
                                                           ess_tdof,
                                                           2)
                tmp_cols.append(opr.shape[0])
                tmp_cols.append(opr.shape[0])
                tmp_diags.append(rsmoother)
                tmp_diags.append(ismoother)

            else:
                dd = opr.diagonal()
                diag = mfem.Vector(list(dd))
                smoother = mfem.OperatorChebyshevSmoother(mat1,
                                                          diag,
                                                          ess_tdof,
                                                          2)
                tmp_diags.append(smoother)
                tmp_cols.append(opr.shape[0])

        else:
            tmp_cols.append(widths[offset])
            tmp_diags.append(mfem.IdentityOperator(widths[offset]))
            if A.complex:
                tmp_cols.append(widths[offset])
                oo = mfem.IdentityOperator(widths[offset])
                if conv == -1:
                    oo2 = mfem.ScaleOperator(oo, -1)
                    oo2._opr = oo
                    tmp_diags.append(oo2)
                else:
                    tmp_diags.append(oo)

        if use_complex_opr:
            tmp_cols = [0] + tmp_cols
            blockOffsets = mfem.intArray(tmp_cols)
            blockOffsets.PartialSum()
            smoother = mfem.BlockDiagonalPreconditioner(blockOffsets)
            smoother.SetDiagonalBlock(0, tmp_diags[0])
            smoother.SetDiagonalBlock(1, tmp_diags[1])
            smoother._smoother = tmp_diags
            cols.append(tmp_cols[1]*2)
            diags.append(smoother)
        else:
            cols.extend(tmp_cols)
            diags.extend(tmp_diags)

    co = mfem.intArray(cols)
    co.PartialSum()

    P = mfem.BlockDiagonalPreconditioner(co)
    for i, d in enumerate(diags):
        P.SetDiagonalBlock(i,  d)
    P._diags = diags
    P._ess_tdofs = ess_tdofs
    return P



from petram.mfem_config import use_parallel
if use_parallel:
    from petram.helper.mpi_recipes import *
    import mfem.par as mfem
else:
    import mfem.ser as mfem

import numpy as np


class HierarchicalFiniteElementSpaces(object):
    '''
    obj.fespaces[name] : returns fes using level_idx

    fec, fes, P and refined_mesh are stored interally for reuse
    '''

    def __init__(self, owner=None):
        self._owner = owner
        self._hierarchies = {}
        self._dataset = {}
        self._fec_storage = {}
        self._fes_storage = {}
        self._p_storage = {}
        self._refined_mesh_storage = {}

    def __getitem__(self, name):
        if not name in self._dataset:
            raise KeyError()
        d = self._dataset[name]
        key1 = d[self._owner.level_idx]
        fes1 = self.get_or_allocate_fes(*key1)
        return fes1

    def __iter__(self):
        return self._dataset.__iter__()

    @property
    def hierarchies(self):
        self._hierarchies = {}

    def get_hierarchy(self, name):
        return self._hierarchies[name]

    def new_hierarchy(self, name, parameters=None):

        emesh_idx, element, order, fecdim, vdim = parameters

        if not (emesh_idx, 0) in self._refined_mesh_storage:
            self._refined_mesh_storage[(
                emesh_idx, 0)] = self._owner.emeshes[emesh_idx]

        mesh = self._refined_mesh_storage[(emesh_idx, 0)]
        mesh.GetEdgeVertexTable()

        fec = self.get_or_allocate_fec(element, order, fecdim)
        fes = self.get_or_allocate_fes(
            emesh_idx, 0, element, order, fecdim, vdim)

        h = self._owner.new_fespace_hierarchy(mesh, fes, False, False)

        self._hierarchies[name] = h
        self._dataset[name] = []
        self._dataset[name].append(
            (emesh_idx, 0, element, order, fecdim, vdim))

    def get_or_allocate_fec(self, element, order, fecdim):
        entry = (element, order, fecdim)
        if entry in self._fec_storage:
            return self._fec_storage[entry]
        else:
            fec_class = getattr(mfem, element)
            fec = fec_class(order, fecdim)
            self._fec_storage[entry] = fec
            return fec

    def get_or_allocate_fes(self, emesh_idx, refine, element, order, fecdim, vdim):
        entry = (emesh_idx, refine, element, order, fecdim, vdim)
        mesh = self._refined_mesh_storage[(emesh_idx, refine)]

        if entry in self._fes_storage:
            return self._fes_storage[entry]
        else:
            fec = self.get_or_allocate_fec(element, order, fecdim)
            fes = self._owner.new_fespace(mesh, fec, vdim)
            self._fes_storage[entry] = fes
            return fes

    def get_or_allocate_transfer(self, name, key1, key2, engine):
        if (key1, key2) in self._p_storage:
            return self._p_storage[(key1, key2)]
        else:
            P = self._new_transfer_operator(name, key1, key2, engine)
            self._p_storage[(key1, key2)] = P
        return P

    def add_same_level(self, name, engine):
        emesh_idx, old_refine, element, order, fecdim, vdim = self._dataset[name][-1]

        m2 = self._refined_mesh_storage[(emesh_idx, old_refine)]
        fec = self.get_or_allocate_fec(element, order, fecdim)

        key1 = (emesh_idx, old_refine, element, order, fecdim, vdim)

        fes1 = self.get_or_allocate_fes(*key1)
        P = self.get_or_allocate_transfer(name, key1, key1, engine)

        h = self._hierarchies[name]
        h.AddLevel(m2, fes1, P, False, False, False)

        self._dataset[name].append(key1)
        return len(self._dataset[name])

    def add_mesh_refined_level(self, name, engine, inc=1, refine_dom=None):
        emesh_idx, old_refine, element, order, fecdim, vdim = self._dataset[name][-1]

        new_refine = old_refine+1
        if not (emesh_idx, new_refine) in self._refined_mesh_storage:
            m = self._refined_mesh_storage[(emesh_idx, old_refine)]
            m2 = self._owner.new_mesh_from_mesh(m)
            for i in range(inc):
                if refine_dom is None:
                    m2.UniformRefinement()
                else:
                    attr = m2.GetAttributeArray()
                    idx = list(np.where(np.in1d(attr, refine_dom))[0])
                    idx0 = mfem.intArray(idx)
                    m2.GeneralRefinement(idx0)  # this is parallel refinement
            m2.GetEdgeVertexTable()
            self._refined_mesh_storage[(emesh_idx, old_refine+inc)] = m2
        else:
            m2 = self._refined_mesh_storage[(emesh_idx, new_refine)]

        fec = self.get_or_allocate_fec(element, order, fecdim)

        key1 = (emesh_idx, old_refine, element, order, fecdim, vdim)
        key2 = (emesh_idx, new_refine, element, order, fecdim, vdim)

        fes1 = self.get_or_allocate_fes(*key1)
        fes2 = self.get_or_allocate_fes(*key2)
        P = self.get_or_allocate_transfer(name, key1, key2, engine)

        h = self._hierarchies[name]
        h.AddLevel(m2, fes2, P, False, False, False)

        self._dataset[name].append(key2)
        return len(self._dataset[name])

    def add_order_refined_level(self, name, engine, inc=1):
        emesh_idx, old_refine, element, order, fecdim, vdim = self._dataset[name][-1]

        m = self._refined_mesh_storage[(emesh_idx, old_refine)]

        key1 = (emesh_idx, old_refine, element, order, fecdim, vdim)
        key2 = (emesh_idx, old_refine, element, order+inc, fecdim, vdim)

        fes1 = self.get_or_allocate_fes(*key1)
        fes2 = self.get_or_allocate_fes(*key2)
        P = self.get_or_allocate_transfer(name, key1, key2, engine)

        h = self._hierarchies[name]
        h.AddLevel(m, fes2, P, False, False, False)

        self._dataset[name].append(key2)
        return len(self._dataset[name])

    def get_fes_info(self, fes):
        for key in self._fes_storage:
            if self._fes_storage[key] == fes:
                emesh_idx, refine, element, order, fecdim, vdim = key
                if hasattr(fes, 'GroupComm'):
                    m = fes.GetParMesh()
                else:
                    m = fes.GetMesh()
                return {'emesh_idx': emesh_idx,
                        'refine': refine,
                        'element': element,
                        'order': order,
                        'dim': m.Dimension(),
                        'sdim': m.SpaceDimension(),
                        'vdim': vdim,
                        'fecdim': fecdim}
        return None

    def get_fes_from_info(self, info):
        emesh_idx = info['emesh_idx']
        refine = info['refine']
        element = info['element']
        order = info['order']
        vdim = info['vdim']
        fecdim = info['fecdim']
        key = emesh_idx, refine, element, order, fecdim, vdim
        return self._fes_storage[key]

    def get_fes_emesh_idx(self, fes):
        info = self.get_fes_info(fes)
        if info is not None:
            return info['emesh_idx']
        else:
            return None

    def get_fes_levels(self, name):
        return len(self._dataset[name])

    def get_mesh(self, name, ref_level=0):
        emesh_idx, refine, element, order, fecdim, vdim = self._dataset[name][-1]
        m = self._refined_mesh_storage[(emesh_idx, refine-ref_level)]
        #m = self._refined_mesh_storage[(emesh_idx, 0)]
        return m

    def _new_transfer_operator(self, name, key1, key2, engine, use_matrix_free=True):
        '''
        fes1 : coarse grid
        fes2 : fine grid
        '''
        fes1 = self._fes_storage[key1]
        fes2 = self._fes_storage[key2]

        parallel = hasattr(fes1, 'GroupComm')
        if use_matrix_free:
            if key1 == key2:
                return mfem.IdentityOperator(fes1.GetTrueVSize())
            elif parallel:
                return mfem.TrueTransferOperator(fes1, fes2)
            else:

                P1 = None if fes1.Conforming() else fes1.GetConformingProlongation()
                if P1 is None:
                    P1 = mfem.IdentityOperator(fes1.GetTrueVSize())

                R2 = None if fes2.Conforming() else fes2.GetConformingRestriction()
                if R2 is None:
                    R2 = mfem.IdentityOperator(fes2.GetTrueVSize())

                Opr1 = mfem.TransferOperator(fes1, fes2)
                opr = mfem.TripleProductOperator(
                    R2, Opr1, P1, False, False, False)
                opr._linked_opr = (R2, Opr1, P1)
                return opr

        else:
            if parallel:
                PP = mfem.OperatorPtr(mfem.Operator.Hypre_ParCSR)

                fes2.GetTransferOperator(fes1, PP)
                PP.SetOperatorOwner(False)
                P = PP.Ptr()
                return P
            else:
                assert False, "Should not come here"



import traceback

import petram.debug
dprint1, dprint2, dprint3 = petram.debug.init_dprints('Projection(PP)')

from petram.phys.vtable import VtableElement, Vtable, Vtable_mixin


from petram.postprocess.pp_model import PostProcessBase

data = [("coeff_lambda", VtableElement("coeff_lambda", type='array',
                                       guilabel = "expression",
                                       default = 0.0,
                                       tip = "expression",))]

class DerivedValue(PostProcessBase, Vtable_mixin):
    has_2nd_panel = True
    vt_coeff = Vtable(data)
    
    def attribute_set(self, v):
        v = super(DerivedValue, self).attribute_set(v)
        v["projection_name"] = 'derived'
        v["element"] = 'H1_FECollection'
        v["is_complex_valued"] = False
        v["order"]   = 1
        v["sdim"] = 2
        v['sel_index'] = ['all']
        v['sel_index_txt'] = 'all'
        self.vt_coeff.attribute_set(v)        
        return v
    
    @property
    def vdim(self):
        name = [x.strip() for x in self.projection_name.split(',')]

        if (self.element.startswith('ND') or
            self.element.startswith('RT')):
            sdim = self.geom_dim
        else:
            sdim = 1
        return len(name)*sdim

    @property
    def geom_dim(self):  # dim of geometry
        return self.root()['Mesh'].sdim
    
    def panel1_param(self):
        import wx
        pa = self.vt_coeff.panel_param(self)        
        panels  =[["new variable name", "", 0, {}],
                  pa[0],
                  ["element", "H1_FECollection", 4,
                                 {"style":wx.CB_READONLY, 
                                   "choices": ["H1_FECollection",
                                               "L2_FECollection",
                                               "ND_FECollection",
                                               "RT_FECollection",
                                               "DG_FECollection"]}],
                  ["order",     self.order,     400,  {}],
                  ["complex", self.is_complex_valued, 3, {"text":""}],]

        return panels

    def panel1_tip(self):
        return ["name", "expression", "element type", "element order", "complex"]
     
    def get_panel1_value(self):                
        return [self.projection_name,
                self.vt_coeff.get_panel_value(self)[0],
                self.element, self.order, self.is_complex_valued]
    
    def import_panel1_value(self, v):
        self.projection_name = str(v[0])
        self.vt_coeff.import_panel_value(self, (v[1],))        
        self.element = str(v[2])
        self.order = int(v[3])
        self.is_complex_valued = bool(v[4])

    def panel2_param(self):
        import wx
        
        if self.geom_dim == 3:
           choice = ("Volume", "Surface", "Edge")
        elif self.geom_dim == 2:
           choice = ("Surface", "Edge")
        elif self.geom_dim == 1:
           choice = ("Edge", )

        p = ["Type", choice[0], 4,
             {"style":wx.CB_READONLY, "choices": choice}]
        return [p, ["index",  'all',  0,   {'changing_event':True,
                                            'setfocus_event':True}, ]]
              
    def get_panel2_value(self):
        choice = ["Point", "Edge", "Surface", "Volume",]
        return choice[self.sdim], self.sel_index_txt
     
    def import_panel2_value(self, v):
        if str(v[0]) == "Volume":
           self.sdim = 3
        elif str(v[0]) == "Surface":
           self.sdim = 2
        elif str(v[0]) == "Edge":
           self.sdim = 1                      
        else:
           self.sdim = 1                                 
        self.sel_index_txt = str(v[1])
           
        from petram.model import convert_sel_txt
        try:
            g = self._global_ns            
            arr = convert_sel_txt(self.sel_index_txt, g)
            self.sel_index = arr            
        except:
            import traceback
            traceback.print_exc()
            assert False, "failed to convert "+self.sel_index_txt

    def get_emesh_idx(self, engine):
        
        from petram.mesh.mesh_extension import MeshExtInfo
        
        info = MeshExtInfo(dim = self.sdim, base = 0)
        if self.sel_index[0] != 'all':        
           info.set_selection(self.sel_index)
        idx = engine.emesh_data.add_info(info)
        
        if len(engine.emeshes) <= idx:
            assert False, "Extended Mesh was not generated for this postprocessing"
        return idx
            
    def run_postprocess(self, engine):
        dprint1("running postprocess: " + self.name())

        names = [x.strip() for x in self.projection_name.split(',')]
        
        emesh_idx = self.get_emesh_idx(engine)
        sdim= self.geom_dim

        if (self.element.startswith('ND') or
            self.element.startswith('RT')):
            vdim = self.vdim//self.geom_dim
            #coeff_dim = self.geom_dim
        else:
            vdim = self.vdim
            #coeff_dim = vdim

        _is_new, fes = engine.get_or_allocate_fecfes(''.join(names),
                                                         emesh_idx,
                                                         self.element,
                                                         self.order,
                                                         vdim)
        gfr = engine.new_gf(fes)
        if self.is_complex_valued:
            gfi = engine.new_gf(fes)
        else:
            gfi = None


        self.vt_coeff.preprocess_params(self)
        c = self.vt_coeff.make_value_or_expression(self)

        phys = self.root()['Phys'].values()
        for p in phys:
            if p.enabled:
                ind_vars = p.ind_vars
                break
        else:
            assert False, "no phys is enabled"
        from petram.helper.variables import var_g
        global_ns = self._global_ns.copy()
        for k in engine.model._variables:
            global_ns[k] = engine.model._variables[k]
        local_ns = {}
        
        from petram.helper.variables import project_variable_to_gf
        
        project_variable_to_gf(c[0], ind_vars, gfr, gfi,
                               global_ns=global_ns, local_ns = local_ns)

        '''                       
        def project_coeff(gf, coeff_dim, c, ind_vars, real):
            if coeff_dim > 1:
                 #print("vector coeff", c)
                 coeff = VCoeff(coeff_dim, c[0], ind_vars,
                                local_ns, global_ns, real = real)
            else:
                 #print("coeff", c)                
                 coeff = SCoeff(c[0], ind_vars,
                                local_ns, global_ns, real = real)
            gf.ProjectCoefficient(coeff)
            
        project_coeff(gfr, coeff_dim, c, ind_vars, True)
        if gfi is not None:
            project_coeff(gfi, coeff_dim, c, ind_vars, False)
        '''
                               
        from petram.helper.variables import Variables
        v = Variables()
            
        self.add_variables(v, names, gfr, gfi)

        engine.add_PP_to_NS(v)
        engine.save_solfile_fespace(''.join(names), emesh_idx, gfr, gfi)
        
    def add_variables(self, v, names, gfr, gfi):
        for phys_name in self.root()['Phys']:
            if not self.root()['Phys'][phys_name].enabled: continue
            ind_vars = self.root()['Phys'][phys_name].ind_vars

        ind_vars = [x.strip() for x in ind_vars.split(',') if x.strip() != '']

        if gfr is not None:
           fes = gfr.FESpace()
        else:
           fes = gfi.FESpace()
        mesh = fes.GetMesh()
        
        isVector = False; isNormal=False
        if (self.element.startswith('RT') and mesh.Dimension() == 3):
            isVector = True
        if (self.element.startswith('RT') and mesh.Dimension() < 3):
            isNormal = True
        if self.element.startswith('ND'): 
            isVector = True

        from petram.helper.variables import add_scalar
        from petram.helper.variables import add_components
        from petram.helper.variables import GFScalarVariable        

        if isVector:
            add_components(v, names[0], "", ind_vars, gfr, gfi)
        elif isNormal:
            add_scalar(v, names[0]+"n", "", ind_vars, gfr, gfi)                
        else:
            if self.vdim == 1:
                add_scalar(v, names[0], "", ind_vars, gfr, gfi)
            else:
                for k, n in enumerate(names):
                    v[n] = GFScalarVariable(gfr, gfi, comp=k+1)
       
        
    def soldict_to_solvars(self, soldict, variables):

        suffix = ""
        names = [x.strip() for x in self.projection_name.split(',')]
        fname = ''.join(names)
        
        for k in soldict:
            n = '_'.join(k.split('_')[:-1])
            if n == fname:
               sol = soldict[k]
               solr = sol[0]
               soli = sol[1] if len(sol) > 1 else None
               self.add_variables(variables, names, solr, soli)
               
            
class DerivedBdrValue(PostProcessBase, Vtable_mixin):
    pass


    




'''

  * Utility for memory_report

  >>>  from petram.helper.memory_report import read_report
  >>>  mem_out, swap_out = read_report()
  or
  >>> mem_out, swap_out = read_report(file='')
  >>> figure();plot(mem_out['used'])

'''
import os 
from petram.mesh.nastran2mfem import NASReader, write_nas2mfem
from os.path import expanduser
import numpy as np
def str2number(s):
    s=s.replace('B', '')    
    s=s.replace('G', '*1e9')
    s=s.replace('M', '*1e6')
    s=s.replace('K', '*1e3')
    return eval(s)


def read_report(file = None, offset=None):
    if file is None:
        from ifigure.widgets.dialog import read
        path = read(message='Memory Report file to read', defaultdir=os.getcwd())
        if path == '': return
    else:
        path = expanduser(file)

    fid = open(path);lines=fid.readlines();fid.close()

    keys = lines[0].strip().split()
    
    mem_out = []
    swp_out = []

    for ll in lines[1:]:
        txts =ll.strip().split()
        if txts[0].startswith('Mem'):
            mem_out.append([str2number(s) for s in txts[1:]])
        if txts[0].startswith('Swap'):
            swp_out.append([str2number(s) for s in txts[1:]])
 
    mem_out = np.array(mem_out).transpose()
    swp_out = np.array(swp_out).transpose()
    
    mem_out = {keys[i]:mem_out[i] for i in range(len(mem_out))}
    swp_out = {keys[i]:swp_out[i] for i in range(len(swp_out))}

    return mem_out, swp_out 




import numpy as np

import petram.debug
from petram.mfem_config import use_parallel

if use_parallel:
    import mfem.par as mfem
    from mpi4py import MPI
    num_proc = MPI.COMM_WORLD.size
    myid = MPI.COMM_WORLD.rank
    from petram.helper.mpi_recipes import *
else:
    import mfem.ser as mfem

dprint1, dprint2, dprint3 = petram.debug.init_dprints('BoundaryRefinement')



def apply_boundary_refinement(mesh, sels, nlayers=6):
    '''
    refinement near the boundary element sels.
    '''
    dprint1("applying boundary refinement: " + str(sels) + ", nlayers=" + str(nlayers))
    dim = mesh.Dimension()

    bdrs = mesh.GetBdrAttributeArray()
    ibdrs = np.where(np.in1d(bdrs, sels))[0]

    ifaces = [mesh.GetBdrFace(i) for i in ibdrs]
    iels = [mesh.GetFaceElementTransformations(i).Elem1No for i in ifaces]

    layers = [iels]
    iels = layers[-1][:]

    if dim == 2:
        for i in range(nlayers):
            new_layer = []
            for i in layers[-1]:
                edges = mesh.GetElementEdges(i)[0]
                for edge in edges:
                    trs = mesh.GetFaceElementTransformations(edge)
                    e1 = trs.Elem1No
                    e2 = trs.Elem2No
                    if e1 not in iels and e1 != -1:
                        if e1 not in new_layer:
                            new_layer.append(e1)
                    if e2 not in iels and e2 != -1:
                        if e2 not in new_layer:
                            new_layer.append(e2)

            iels.extend(new_layer)
            layers.append(new_layer)

    elif dim == 3:
        for i in range(nlayers):
            new_layer = []
            for i in layers[-1]:
                faces = mesh.GetElementFaces(i)[0]
                for face in faces:
                    trs = mesh.GetFaceElementTransformations(face)
                    e1 = trs.Elem1No
                    e2 = trs.Elem2No
                    if e1 not in iels and e1 != -1:
                        if e1 not in new_layer:
                            new_layer.append(e1)
                    if e2 not in iels and e2 != -1:
                        if e2 not in new_layer:
                            new_layer.append(e2)

            iels.extend(new_layer)
            layers.append(new_layer)

        pass

    else:
        pass

    iii = mfem.intArray(iels)
    mesh.GeneralRefinement(iii)
    return mesh
